from collections import namedtuple
from datetime import timedelta

from sqlalchemy import types
from sqlalchemy.orm import deferred
from sqlalchemy.schema import Column
from sqlalchemy.schema import Index

from seantis.reservation import ORMBase
from seantis.reservation import Session
from seantis.reservation.models import customtypes
from seantis.reservation.models.other import OtherModels
from seantis.reservation.models.timestamp import TimestampMixin


Timespan = namedtuple(
    'Timespan', ('start', 'end')
)

BoundTimespan = namedtuple(
    'BoundTimespan', ('start', 'end', 'token', 'id')
)


class Reservation(TimestampMixin, ORMBase, OtherModels):
    """Describes a pending or approved reservation.

    """

    __tablename__ = 'reservations'

    id = Column(
        types.Integer(),
        primary_key=True,
        autoincrement=True
    )

    token = Column(
        customtypes.GUID(),
        nullable=False,
    )

    target = Column(
        customtypes.GUID(),
        nullable=False,
    )

    target_type = Column(
        types.Enum(u'group', u'allocation', name='reservation_target_type'),
        nullable=False
    )

    resource = Column(
        customtypes.GUID(),
        nullable=False
    )

    start = Column(
        types.DateTime(),
        nullable=True
    )

    end = Column(
        types.DateTime(),
        nullable=True
    )

    status = Column(
        types.Enum(u'pending', u'approved', name="reservation_status"),
        nullable=False
    )

    data = deferred(
        Column(
            customtypes.JSONEncodedDict(),
            nullable=True
        )
    )

    email = Column(
        types.Unicode(254),
        nullable=False
    )

    session_id = Column(
        customtypes.GUID()
    )

    quota = Column(
        types.Integer(),
        nullable=False
    )

    __table_args__ = (
        Index('target_status_ix', 'status', 'target', 'id'),
    )

    @property
    def display_start(self):
        """Does nothing but to form a nice pair to display_end."""
        return self.start

    @property
    def display_end(self):
        """Returns the end plus one microsecond (nicer display)."""
        return self.end + timedelta(microseconds=1)

    def _target_allocations(self):
        """ Returns the allocations this reservation is targeting. This should
        NOT be confused with db.allocations_by_reservation. The method in
        the db module returns the actual allocations belonging to an approved
        reservation.

        This method only returns the master allocations to get information
        about timespans and other properties. If you don't know exactly
        what you're doing you do not want to use this method as misuse might
        be dangerous.

        """
        Allocation = self.models.Allocation
        query = Session.query(Allocation)
        query = query.filter(Allocation.group == self.target)

        # master allocations only
        query = query.filter(Allocation.resource == Allocation.mirror_of)

        return query

    def timespans(self):

        if self.target_type == u'allocation':
            return [(self.start, self.end + timedelta(microseconds=1))]
        elif self.target_type == u'group':
            return [
                Timespan(
                    a.display_start, a.display_end
                )
                for a in self._target_allocations()
            ]
        else:
            raise NotImplementedError

    def bound_timespans(self):

        return [
            BoundTimespan(
                t[0], t[1], self.token, self.id
            ) for t in self.timespans
        ]

    @property
    def title(self):
        return self.email

    @property
    def autoapprovable(self):
        query = self._target_allocations()
        query = query.filter(self.models.Allocation.approve_manually == True)

        # A reservation is deemed autoapprovable if no allocation
        # requires explicit approval

        return query.first() is None



from datetime import datetime, timedelta

from five import grok
from plone import api
from zope.interface import Interface

from sqlalchemy import desc

from seantis.reservation import Session
from seantis.reservation import db
from seantis.reservation import _
from seantis.reservation import utils
from seantis.reservation.reservations import combine_reservations
from seantis.reservation.models import Reservation
from seantis.reservation.base import BaseView
from seantis.reservation.reports import GeneralReportParametersMixin


def human_date(date):
    # timezones are currently naive and implicity the one used by
    # the users - we don't have international reservations yet
    now = utils.utcnow()

    this_morning = datetime(
        now.year, now.month, now.day
    ).replace(tzinfo=now.tzinfo)

    # remove the timezone before getting the localized time - or plone.api
    # will adjust it.. it's really weird
    time = api.portal.get_localized_time(
        date.replace(tzinfo=None), time_only=True
    )

    if date >= this_morning:
        return _(u'Today, at ${time}', mapping={'time': time})

    days = (now.date() - date.date()).days

    if days <= 1:
        return _(u'Yesterday, at ${time}', mapping={
            'time': time
        })
    else:
        return _(u'${days} days ago, at ${time}', mapping={
            'days': days,
            'time': time
        })


class LatestReservationsReportView(BaseView, GeneralReportParametersMixin):

    permission = 'seantis.reservation.ViewReservations'

    grok.require(permission)

    grok.context(Interface)
    grok.name('latest_reservations')

    template = grok.PageTemplateFile('../templates/latest_reservations.pt')

    @property
    def title(self):
        return _(u'Latest Reservations')

    @property
    def start(self):
        return utils.safe_parse_int(self.request.get('start'), 0)

    @property
    def end(self):
        return utils.safe_parse_int(self.request.get('end'), 30)

    @property
    def results(self):
        return latest_reservations(
            resources=self.resources,
            reservations=self.reservations or '*',
            daterange=self.daterange,
        )

    @property
    def daterange(self):
        now = utils.utcnow()

        since = now - timedelta(days=self.end)
        until = now - timedelta(days=self.start)

        return since, until

    @property
    def daterange_label(self):
        since, until = self.daterange

        if until.date() == utils.utcnow().date():
            return ' - '.join((
                api.portal.get_localized_time(since, long_format=False),
                self.translate(_(u'Today'))
            ))
        return ' - '.join((
            api.portal.get_localized_time(since, long_format=False),
            api.portal.get_localized_time(until, long_format=False)
        ))

    def build_url(self, start, end):
        params = [
            ('start', str(start)),
            ('end', str(end))
        ]

        return super(LatestReservationsReportView, self).build_url(
            extra_parameters=params
        )

    @property
    def forward_url(self):
        start, end = self.start, self.end
        start, end = start - 30, end - 30

        if start < 0:
            return None

        return self.build_url(start, end)

    @property
    def backward_url(self):
        start, end = self.start, self.end
        start, end = start + 30, end + 30

        return self.build_url(start, end)

    def reservation_title(self, reservation):
        human_date_text = utils.translate(
            self.context, self.request, human_date(reservation.created)
        )
        return '{} - {}'.format(human_date_text, reservation.title)

    def unique(self, reservations):
        return tuple(combine_reservations(reservations))


def latest_reservations(resources, daterange, reservations='*'):
    schedulers = {}

    for uuid in resources.keys():
        schedulers[uuid] = db.Scheduler(uuid)

    query = Session.query(Reservation)
    query = query.filter(Reservation.resource.in_(resources.keys()))
    query = query.filter(Reservation.created > daterange[0])
    query = query.filter(Reservation.created <= daterange[1])
    query = query.order_by(desc(Reservation.created))

    if reservations != '*':
        query = query.filter(Reservation.token.in_(reservations))

    result = utils.OrderedDict()
    for reservation in query.all():
        if reservation.token in result:
            result[reservation.token].append(reservation)
        else:
            result[reservation.token] = [reservation]

    return result



import json
import transaction

from datetime import timedelta, datetime
from itertools import chain

from zope.component import queryUtility
from Products.CMFCore.interfaces import IPropertiesTool

from seantis.reservation import utils
from seantis.reservation import db
from seantis.reservation.tests import FunctionalTestCase


class FormsetField(object):
    def __init__(self, name, type, default=None):
        self.name = name
        self.type = type
        self.default = default


class TestBrowser(FunctionalTestCase):

    def setUp(self):
        super(TestBrowser, self).setUp()
        self.baseurl = self.portal.absolute_url()

        browser = self.new_browser()
        browser.login_admin()

        # create a testfolder for each test
        browser.open(self.baseurl + '/createObject?type_name=Folder')

        browser.getControl('Title').value = 'testfolder'
        browser.getControl('Save').click()

        self.assertTrue('no items in this folder' in browser.contents)

        self.admin_browser = browser
        self.folder_url = self.baseurl + '/testfolder'
        self.infolder = lambda url: self.folder_url + url
        self.formsets = []

    def tearDown(self):
        browser = self.admin_browser

        # delete the destfolder again
        browser.open(self.infolder('/delete_confirmation'))
        browser.getControl('Delete').click()

        browser.assert_notfound(self.baseurl + '/testfolder')

        # delete the formsests
        if self.formsets:
            browser.open('/dexterity-types')

            for formset in self.formsets:
                input_name = 'crud-edit.{}.widgets.select:list'.format(formset)
                browser.getControl(name=input_name).value = [True]

            browser.getControl('Delete').click()

            for formset in self.formsets:
                browser.assert_notfound('/dexterity-types/{}'.format(formset))

        super(TestBrowser, self).tearDown()

    def add_resource(
        self, name, description='', formsets=[], action=None, thank_you=''
    ):
        browser = self.admin_browser

        browser.open(self.infolder('/++add++seantis.reservation.resource'))
        browser.getControl('Name').value = name
        browser.getControl('Description').value = description
        browser.getControl('Thank you text').value = thank_you

        for formset in formsets:
            browser.getControl(formset).selected = True

        browser.getControl('Save').click()

        if action:
            browser.open(self.infolder(
                '/{}/content_status_modify?workflow_action={}'.format(
                    name, action
                )
            ))

    def add_allocation(
        self, resource, start, end,
        partly_available=False,
        quota=1,
        quota_limit=1,
        approve_manually=False,
        recurrence_start=None,
        recurrence_end=None,
        separately_reservable=False
    ):

        browser = self.admin_browser

        s = utils.utctimestamp(start)
        e = utils.utctimestamp(end)

        allocate_url = '/%s/allocate?start=%s&end=%s' % (resource, s, e)
        browser.open(self.infolder(allocate_url))

        # Plone formats the English run tests like this: 1:30 AM
        # Python does not do single digit hours so we strip.
        ds = start.strftime('%I:%M %p').lstrip('0')
        de = end.strftime('%I:%M %p').lstrip('0')

        self.assertTrue(ds in browser.contents)
        self.assertTrue(de in browser.contents)

        browser.getControl('Partly available').selected = partly_available
        browser.getControl('Quota', index=0).value = str(quota)
        browser.getControl('Reservation Quota Limit').value = str(quota_limit)

        if recurrence_start and recurrence_end:
            browser.getControl('Daily').selected = True
            browser.set_date('form.widgets.recurrence_start', recurrence_start)
            browser.set_date('form.widgets.recurrence_end', recurrence_end)

            for day in ('Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa', 'Su'):
                browser.getControl(day).selected = True

            (
                browser.getControl('Separately reservable')
            ).selected = separately_reservable

        browser.getControl(
            'Manually approve reservation requests'
        ).selected = approve_manually

        browser.getControl('Allocate').click()

        return [resource, start, end]

    def add_formset(self, name, fields, for_managers=False):
        browser = self.admin_browser

        browser.open('/dexterity-types/@@add-type')
        browser.getControl('Type Name').value = name
        browser.getControl('Short Name').value = name

        browser.getControl('Add').click()

        self.formsets.append(name)

        for field in fields:

            browser.open('/dexterity-types/{}/@@add-field'.format(name))
            browser.getControl('Title').value = field.name
            browser.getControl('Short Name').value = field.name
            browser.getControl('Field type').value = (field.type, )
            browser.getControl('Add').click()

            # no requirements for now, there are issues with defaults
            browser.open('/dexterity-types/{}/{}'.format(name, field.name))
            browser.getControl('Required').selected = False
            browser.getControl('Save').click()

        browser.open('/dexterity-types/{}/fields'.format(name))

        for field in (f for f in fields if f.default is not None):

            ctl = browser.getControl(name='form.widgets.{}'.format(field.name))
            ctl.value = field.default

        browser.getControl('Save Defaults').click()

        browser.open('/dexterity-types/{}/@@behaviors'.format(name))

        standard = 'Reservation Formset'
        managers = 'Reservation Manager Formset'

        browser.getControl(standard).selected = not for_managers
        browser.getControl(managers).selected = for_managers

        browser.getControl('Save').click()

    def load_slot_data(self, resource, start, end):

        browser = self.admin_browser

        s = utils.utctimestamp(start)
        e = utils.utctimestamp(end)

        slots_url = '/%s/slots?start=%s&end=%s' % (resource, s, e)

        browser.open(self.infolder(slots_url))
        return json.loads(browser.contents.replace('\n', ''))

    def allocation_menu(self, resource, start, end):

        slots = self.load_slot_data(resource, start, end)
        assert len(slots) == 1

        menu = {}
        for group, entries in slots[0]['menu'].items():
            for e in entries:
                menu[e['name'].lower()] = e['url'].replace('/plone', '')

        return menu

    def test_resource_listing(self):

        browser = self.new_browser()
        browser.login_admin()

        browser.open(
            self.infolder('/selectViewTemplate?templateId=resource_listing')
        )

        self.add_resource('Resource 1', 'Description 1')

        browser.open(self.folder_url)
        self.assertEqual(browser.contents.count('Click to reserve'), 1)
        self.assertTrue('Resource 1' in browser.contents)
        self.assertTrue('Description 1' in browser.contents)

        self.add_resource('Resource 2')

        browser.open(self.folder_url)
        self.assertEqual(browser.contents.count('Click to reserve'), 2)
        self.assertTrue('Resource 2' in browser.contents)

    def test_singlecalendar(self):

        browser = self.new_browser()
        browser.login_admin()

        self.add_resource('Test')

        browser.open(self.infolder('/test'))

        self.assertTrue('singlecalendar' in browser.contents)
        self.assertFalse('multicalendar' in browser.contents)

        self.assertTrue('test' in browser.contents)

        # the title of the parent is not used as a prefix in this case
        self.assertFalse('testfolder - test' in browser.contents)

    def test_multicalendar(self):

        browser = self.new_browser()
        browser.login_admin()

        self.add_resource('one')
        self.add_resource('two')

        browser.open(self.infolder('/one/@@uuid'))
        uuid = browser.contents

        browser.open(self.infolder('/two?compare_to=%s' % uuid))

        self.assertFalse('singlecalendar' in browser.contents)
        self.assertTrue('multicalendar' in browser.contents)

        # the title of the parent is used as a prefix
        self.assertTrue('testfolder - one' in browser.contents)
        self.assertTrue('testfolder - two' in browser.contents)

    def test_resource_properties(self):

        browser = self.new_browser()
        browser.login_admin()

        self.add_resource('test-resource')

        browser.open(self.infolder('/test-resource/@@edit'))
        browser.getControl('First hour of the day').value = "10"
        browser.getControl('Last hour of the day').value = "12"

        browser.getControl(name='form.widgets.available_views:list').value = [
            'month', 'agendaWeek'
        ]

        try:     # plone 4.3
            control = 'form.widgets.selected_view'
            browser.getControl(name=control).value = ['month']
        except:  # plone 4.2
            control = 'form.widgets.selected_view:list'
            browser.getControl(name=control).value = ['month']

        browser.getControl('Save').click()

        self.assertTrue('"minTime": 10' in browser.contents)
        self.assertTrue('"maxTime": 12' in browser.contents)
        self.assertTrue('"right": "month, agendaWeek"' in browser.contents)
        self.assertTrue('"defaultView": "month"' in browser.contents)

    def test_render_disabled_dates_partly(self):

        # ensure that the disabled date/time is rendered again after
        # submitting a form

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 9, 20, 15, 0)
        end = datetime(2013, 9, 20, 16, 0)

        self.add_resource('render')

        allocation = ('render', start, end)
        self.add_allocation(*allocation, partly_available=True)

        browser.open(self.allocation_menu(*allocation)['reserve'])

        unchanging_values = [
            ('#form-widgets-day-day', '20'),
            ('#form-widgets-day-month option[selected="selected"]', '9'),
            ('#form-widgets-day-year', '2013'),
        ]

        changing_values = [
            ('#form-widgets-start_time', '3:00 PM'),
            ('#form-widgets-end_time', '4:00 PM')
        ]

        for selector, value in chain(unchanging_values, changing_values):
            self.assertEqual(browser.query(selector).val(), value)

        browser.getControl('Start').value = '3:15 PM'
        browser.getControl('End').value = '3:45 PM'

        # not entering anything should lead to the form again with the still
        # filled out controls
        browser.getControl('Reserve').click()

        changing_values = [
            ('#form-widgets-start_time', '3:15 PM'),
            ('#form-widgets-end_time', '3:45 PM')
        ]

        for selector, value in chain(unchanging_values, changing_values):
            self.assertEqual(browser.query(selector).val(), value)

    def test_render_disabled_dates_non_partly(self):

        # ensure that the disabled date/time is rendered again after
        # submitting a form (on not partly_available allocations)

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 9, 20, 15, 0)
        end = datetime(2013, 9, 20, 16, 0)

        self.add_resource('render')

        allocation = ('render', start, end)
        self.add_allocation(*allocation, partly_available=False)

        browser.open(self.allocation_menu(*allocation)['reserve'])

        unchanging_values = [
            ('#form-widgets-day-day', '20'),
            ('#form-widgets-day-month option[selected="selected"]', '9'),
            ('#form-widgets-day-year', '2013'),
            ('#form-widgets-start_time', '3:00 PM'),
            ('#form-widgets-end_time', '4:00 PM')
        ]

        for selector, value in unchanging_values:
            self.assertEqual(browser.query(selector).val(), value)

        # not entering anything should lead to the form again with the still
        # filled out controls
        browser.getControl('Reserve').click()

        for selector, value in unchanging_values:
            self.assertEqual(browser.query(selector).val(), value)

    def test_invalid_allocation_missing_email_regression(self):

        # if an email is entered but the reservation is invalid the email
        # may not disappear

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 3, 4, 15, 0)
        end = datetime(2013, 3, 4, 16, 0)

        self.add_resource('regression')

        allocation = ('regression', start, end)
        self.add_allocation(*allocation, partly_available=True)

        browser.open(self.allocation_menu(*allocation)['reserve'])

        # enter some invalid dates and an email. The form should then show an
        # error, but not forget about the email like it used to
        browser.getControl('End').value = '1:00 PM'
        browser.getControl('Email').value = 'test@example.com'

        browser.getControl('Reserve').click()

        self.assertTrue('Reservation out of bounds' in browser.contents)
        self.assertTrue('test@example.com' in browser.contents)

        # really reserve this time
        browser.getControl('End').value = '4:00 PM'
        browser.getControl('Reserve').click()

        self.assertTrue('Your reservations' in browser.contents)

    def test_over_quota_limit_regression(self):

        # if a quota is reserved which is over the allowed amount the quota
        # must not just be changed, but there should be an error

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 3, 4, 15, 0)
        end = datetime(2013, 3, 4, 16, 0)

        self.add_resource('regression')

        allocation = ('regression', start, end)
        self.add_allocation(*allocation, quota=4, quota_limit=2)
        browser.open(self.allocation_menu(*allocation)['reserve'])

        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Number of Reservations', index=0).value = '3'
        browser.getControl('Reserve').click()
        self.assertTrue(
            'number of reservations is higher than allowed' in browser.contents
        )
        self.assertEqual(
            '3', browser.getControl('Number of Reservations', index=0).value
        )

    def test_overreserve_regression(self):

        # it was possible to add reservations to the waitinglist before,
        # even if there was no waitinglist allowed.
        # it needed a quota to be higher than the free spots

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 6, 2, 14, 0)
        end = datetime(2013, 6, 2, 16, 0)

        self.add_resource('overreserve')
        allocation = ('overreserve', start, end)
        self.add_allocation(*allocation, quota=5, quota_limit=5)

        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Number of Reservations', index=0).value = '3'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Number of Reservations', index=0).value = '3'

        browser.getControl('Reserve').click()
        self.assertTrue(
            'The requested period is no longer available' in browser.contents
        )

        browser.open(self.infolder('/overreserve/your-reservations'))
        self.assertFalse('limitedList' in browser.contents)
        self.assertFalse('your-reservation-quota' in browser.contents)

    def test_throttling(self):

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 6, 2, 14, 0)
        end = datetime(2013, 6, 2, 16, 0)

        self.add_resource('throttled')
        allocation = ('throttled', start, end)
        self.add_allocation(*allocation, quota=5, quota_limit=5)

        browser.open(self.infolder(
            '/throttled/content_status_modify?workflow_action=publish'
        ))

        browser.logout()

        # validation issues don't lead to throttling (missing email here)
        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Number of Reservations', index=0).value = '1'
        browser.getControl('Reserve').click()
        browser.getControl('Reserve').click()
        self.assertFalse('Too many reservations' in browser.contents)

        # real reservations however do
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()
        self.assertFalse('Too many reservations' in browser.contents)

        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Number of Reservations', index=0).value = '1'
        browser.getControl('Reserve').click()
        self.assertTrue('Too many reservations' in browser.contents)

    def test_your_reservations_context_regression(self):

        # calling the 'your-reservations' view did not work outside of
        # the resource context, even though that view should work independently
        # of context

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 9, 13, 11, 0)
        end = datetime(2013, 9, 13, 12, 0)

        self.add_resource('yours')
        allocation = ('yours', start, end)
        self.add_allocation(*allocation, quota=2)

        # it works on the resource context
        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        # but it should also work somewhere else
        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()

        browser.open(self.infolder('/your-reservations'))
        browser.getControl('Submit Reservations').click()

    def test_your_reservations_removal(self):
        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 9, 13, 11, 0)
        end = datetime(2013, 9, 13, 12, 0)

        self.add_resource('removal')
        allocation = ('removal', start, end)
        self.add_allocation(*allocation, quota=2)

        browser.open(self.allocation_menu(*allocation)['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()

        browser.open(self.infolder('/removal'))

        # the datetime in the browsers cannot be swayed, it always is english
        self.assertTrue('Sep 13, 2013 11:00 AM - 12:00 PM' in browser.contents)

        browser.getLink('Remove').click()
        browser.open(self.infolder('/removal'))

        # the datetime in the browsers cannot be swayed, it always is english
        self.assertFalse(
            'Sep 13, 2013 11:00 AM - 12:00 PM' in browser.contents
        )

    def test_reservation_approval(self):

        browser = self.new_browser()
        browser.login_admin()

        start = datetime(2013, 6, 21, 13, 0)
        end = datetime(2013, 6, 21, 17, 0)

        self.add_resource('approval')

        allocation = ('approval', start, end)
        self.add_allocation(*allocation, approve_manually=True)
        menu = self.allocation_menu(*allocation)
        browser.open(menu['reserve'])

        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()

        browser.getControl('Submit Reservations').click()

        browser.open(menu['manage'])

        self.assertTrue('Approve' in browser.contents)
        self.assertTrue('Deny' in browser.contents)

        browser.getLink('Approve').click()

        self.assertTrue('Concerned Dates' in browser.contents)

        # the datetime in the browsers cannot be swayed, it always is english
        self.assertTrue('Jun 21, 2013 01:00 PM - 05:00 PM' in browser.contents)

        browser.getControl('Approve').click()
        browser.open(menu['manage'])

        self.assertTrue('Revoke' in browser.contents)
        self.assertFalse('Approve' in browser.contents)

    def test_reserve_group(self):

        browser = self.admin_browser

        self.add_resource('group_reservation', action='publish')

        start = datetime(2014, 2, 25, 12, 00)
        end = datetime(2014, 2, 25, 15, 00)

        allocation = ['group_reservation', start, end]
        self.add_allocation(
            *allocation,
            recurrence_start=start.date(),
            recurrence_end=start.date() + timedelta(days=3)
        )
        menu = self.allocation_menu(*allocation)

        browser.open(menu['reserve'])

        self.assertEqual(browser.query('.result-time').length, 4)

        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()

        self.assertTrue('Your reservations' in browser.contents)

    def test_reservation_formsets(self):

        browser = self.admin_browser

        self.add_formset(
            'public', [FormsetField('name', 'Text')], for_managers=False
        )

        self.add_formset(
            'private', [
                FormsetField('price', 'Integer', '31415927'),
                FormsetField('comment', 'Text')
            ], for_managers=True
        )

        self.add_resource(
            'formsets', formsets=['public', 'private'], action='publish'
        )

        start = datetime(2013, 6, 27, 13, 37)
        end = datetime(2013, 6, 27, 17, 15)

        allocation = ['formsets', start, end]
        self.add_allocation(*allocation)
        menu = self.allocation_menu(*allocation)

        # admins see all form elements when reserving
        browser.open(menu['reserve'])

        self.assertTrue('public' in browser.contents)
        self.assertTrue('private' in browser.contents)

        # normal users don't
        anonymous = self.new_browser()
        anonymous.open(menu['reserve'])

        self.assertTrue('public' in anonymous.contents)
        self.assertFalse('private' in anonymous.contents)

        # if the normal user does the reservation, the formset defaults are
        # written even for manager formsets
        anonymous.getControl('Email').value = 'test@example.com'
        anonymous.getControl('Reserve').click()
        anonymous.getControl('Submit Reservations').click()

        # this can be verified on the manage-page
        browser.open(menu['manage'])
        self.assertTrue('31415927' in browser.contents)
        self.assertFalse('comment' in browser.contents)

        # the manager can at this point change the values
        browser.getLink('Edit Formdata').click()

        browser.getControl('comment').value = 'no comment'
        browser.getControl('Save').click()

        browser.open(menu['manage'])
        self.assertTrue('no comment' in browser.contents)

    def test_edit_email(self):
        browser = self.admin_browser

        self.add_resource('email')

        start = datetime(2013, 6, 27, 13, 37)
        end = datetime(2013, 6, 27, 17, 15)

        allocation = self.add_allocation('email', start, end)
        menu = self.allocation_menu(*allocation)

        browser.open(menu['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        # this can be verified on the manage-page
        browser.open(menu['manage'])
        self.assertTrue('test@example.com' in browser.contents)

        # the manager can at this point change the values
        browser.getLink('Edit Formdata').click()

        self.assertIn('test@example.com', browser.contents)

        browser.getControl('Email').value = 'invalid'
        browser.getControl('Save').click()

        self.assertIn('Invalid', browser.contents)

        browser.getControl('Email').value = 'asdf@asdf.com'
        browser.getControl('Save').click()

        browser.open(menu['manage'])
        self.assertIn('asdf@asdf.com', browser.contents)
        self.assertNotIn('test@example.com', browser.contents)

    def test_broken_formsets(self):

        browser = self.admin_browser

        self.add_formset(
            'one', [FormsetField('one', 'Text')], for_managers=False
        )
        self.add_formset(
            'two', [FormsetField('two', 'Text')], for_managers=False
        )

        self.add_resource(
            'broken', formsets=['one', 'two'], action='publish'
        )

        start = datetime(2013, 6, 27, 13, 37)
        end = datetime(2013, 6, 27, 17, 15)

        allocation = ['broken', start, end]
        self.add_allocation(*allocation)
        menu = self.allocation_menu(*allocation)

        browser.open(menu['reserve'])
        browser.getControl('Email').value = 'test@example.com'
        browser.getControl('one').value = 'Eins'
        browser.getControl('two').value = 'Zwei'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        browser.open(menu['manage'])
        self.assertTrue('Eins' in browser.contents)
        self.assertTrue('Zwei' in browser.contents)

        browser.open(self.infolder('/broken/edit'))
        browser.getControl('two').selected = False
        browser.getControl('Save').click()

        browser.open(menu['manage'])
        browser.getLink('Edit Formdata').click()
        self.assertTrue('Unchangeable formdata found!' in browser.contents)
        self.assertFalse('<span>Eins</span>' in browser.contents)
        self.assertTrue('<span>Zwei</span>' in browser.contents)

        browser.open(self.infolder('/broken/edit'))
        browser.getControl('two').selected = True
        browser.getControl('Save').click()

        browser.open(menu['manage'])
        browser.getLink('Edit Formdata').click()
        self.assertFalse('Unchangeable formdata found!' in browser.contents)
        self.assertFalse('<span>Eins</span>' in browser.contents)
        self.assertFalse('<span>Zwei</span>' in browser.contents)

    @db.serialized
    def test_resource_removal(self):

        browser = self.admin_browser

        # have one resource with data during the run, to check that there's
        # no bleed over through the other tests

        self.add_resource('keeper')
        browser.open(self.infolder('/keeper/@@uuid'))
        uuid = unicode(browser.contents)

        keeper = db.Scheduler(uuid)

        self.assertEqual(keeper.managed_allocations().count(), 0)
        self.assertEqual(keeper.managed_reservations().count(), 0)
        self.assertEqual(keeper.managed_reserved_slots().count(), 0)

        dates = [
            datetime(2013, 7, 2, 10, 15), datetime(2013, 7, 2, 10, 30)
        ]
        keeper.allocate(dates, raster=15)

        token = keeper.reserve(u'test@example.com', dates)
        keeper.approve_reservations(token)

        transaction.commit()  # delete_confirmation will rollback
                              # dropping all SQL statements

        self.assertEqual(keeper.managed_allocations().count(), 1)
        self.assertEqual(keeper.managed_reservations().count(), 1)
        self.assertEqual(keeper.managed_reserved_slots().count(), 1)

        def run_test():
            self.add_resource('removal')
            browser.open(self.infolder('/removal/@@uuid'))
            uuid = unicode(browser.contents)

            scheduler = db.Scheduler(uuid)

            self.assertEqual(scheduler.managed_allocations().count(), 0)
            self.assertEqual(scheduler.managed_reservations().count(), 0)
            self.assertEqual(scheduler.managed_reserved_slots().count(), 0)

            scheduler.allocate(dates, raster=15)

            token = scheduler.reserve(u'test@example.com', dates)
            scheduler.approve_reservations(token)

            transaction.commit()  # delete_confirmation will rollback
                                  # dropping all SQL statements

            self.assertEqual(scheduler.managed_allocations().count(), 1)
            self.assertEqual(scheduler.managed_reservations().count(), 1)
            self.assertEqual(scheduler.managed_reserved_slots().count(), 1)

            # linkintegrity will cause IObjectRemovedEvent to be fired,
            # but the transaction will be rolled back
            browser.open(self.infolder('/removal/delete_confirmation'))
            self.assertEqual(scheduler.managed_allocations().count(), 1)
            self.assertEqual(scheduler.managed_reservations().count(), 1)
            self.assertEqual(scheduler.managed_reserved_slots().count(), 1)

            browser.getControl('Cancel').click()
            self.assertEqual(scheduler.managed_allocations().count(), 1)
            self.assertEqual(scheduler.managed_reservations().count(), 1)
            self.assertEqual(scheduler.managed_reserved_slots().count(), 1)

            browser.open(self.infolder('/removal/delete_confirmation'))
            browser.getControl('Delete').click()
            self.assertEqual(scheduler.managed_allocations().count(), 0)
            self.assertEqual(scheduler.managed_reservations().count(), 0)
            self.assertEqual(scheduler.managed_reserved_slots().count(), 0)

        # run once with link integrity enabled, once disabled
        ptool = queryUtility(IPropertiesTool)
        props = getattr(ptool, 'site_properties', None)

        props.enable_link_integrity_checks = True
        transaction.commit()
        run_test()

        self.assertEqual(keeper.managed_allocations().count(), 1)
        self.assertEqual(keeper.managed_reservations().count(), 1)
        self.assertEqual(keeper.managed_reserved_slots().count(), 1)

        props.enable_link_integrity_checks = False
        transaction.commit()
        run_test()

        self.assertEqual(keeper.managed_allocations().count(), 1)
        self.assertEqual(keeper.managed_reservations().count(), 1)
        self.assertEqual(keeper.managed_reserved_slots().count(), 1)

    @db.serialized
    def test_export_selection(self):
        browser = self.admin_browser

        self.add_resource('export')

        browser.open(self.infolder('/export/@@uuid'))
        uuid = browser.contents

        browser.open(
            self.infolder('/reservation-exports?uuid={}'.format(uuid))
        )

        self.assertIn('Reservation Export', browser.contents)
        browser.getControl(name='form.buttons.export').click()

        self.assertIn('uuid={}'.format(uuid), browser.url)
        self.assertIn('year=all', browser.url)
        self.assertIn('month=all', browser.url)

    @db.serialized
    def test_thank_you_page(self):
        browser = self.admin_browser

        self.add_resource('one', thank_you='<b>one thanks</b>')
        self.add_resource('two', thank_you='<b>two thanks</b>')
        self.add_resource('three', thank_you='')

        start, end = (
            datetime(2014, 6, 20, 15, 15), datetime(2014, 6, 20, 15, 30)
        )

        def reserve(resource):
            allocation = [resource, start, end]
            self.add_allocation(*allocation)

            menu = self.allocation_menu(*allocation)

            browser.open(menu['reserve'])
            browser.getControl('Email').value = 'test@example.org'
            browser.getControl('Reserve').click()

        map(reserve, ('one', 'two', 'three'))
        browser.getControl('Submit Reservations').click()

        self.assertIn('thank-you-for-reserving', browser.url)

        # one and two have thank you texts, so they show up
        self.assertIn('testfolder - one', browser.contents)
        self.assertIn('<b>one thanks</b>', browser.contents)

        self.assertIn('testfolder - two', browser.contents)
        self.assertIn('<b>two thanks</b>', browser.contents)

        # three doesn't, so it won't show up
        self.assertNotIn('testfolder - three', browser.contents)

    @db.serialized
    def test_pending_timespans(self):
        browser = self.admin_browser

        self.add_resource('manual')

        allocation = self.add_allocation(
            'manual',
            datetime(2014, 2, 25, 12, 00),
            datetime(2014, 2, 25, 15, 00),
            approve_manually=True
        )

        menu = self.allocation_menu(*allocation)

        # reserve this manually approved allocation
        browser.open(menu['reserve'])
        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        # as long as the reservation is not approved, there's no revoke
        # link for the dates in the reservation (because that would
        # require a separate mail message, would confuse the user and
        # would give the admins a way to willy-nilly change the users
        # reservation request which is what he wants approved
        # anyway - not some different thing)
        browser.open(menu['manage'])
        self.assertEqual(len(browser.query('.timespan-actions a')), 0)

        # approve it
        browser.getLink('Approve').click()
        browser.getControl('Approve').click()

        # ensure that the revoke is now being shown
        browser.open(menu['manage'])
        self.assertEqual(len(browser.query('.timespan-actions a')), 1)

    @db.serialized
    def test_remove_group_timespans(self):
        browser = self.admin_browser

        self.add_resource('group')

        start = datetime(2014, 2, 25, 12, 00)
        end = datetime(2014, 2, 25, 15, 00)

        allocation = self.add_allocation(
            'group', start, end,
            recurrence_start=start.date(),
            recurrence_end=start.date() + timedelta(days=3)
        )
        menu = self.allocation_menu(*allocation)

        browser.open(menu['reserve'])
        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        browser.open(menu['manage'])

        # with this group reservation we should see three dates with one
        # revoke link only.
        self.assertEqual(len(browser.query('.timespan-actions a')), 1)
        self.assertEqual(len(browser.query('.timespan-dates')), 4)

        # clicking one will remove all dates
        browser.getLink('Revoke', index=1).click()
        self.assertEqual(len(browser.query('.limitedList > div')), 4)

        browser.getControl('Revoke').click()

        browser.open(menu['manage'])
        self.assertEqual(len(browser.query('.timespan-dates')), 0)

    @db.serialized
    def test_remove_timespans(self):
        browser = self.admin_browser

        self.add_resource('timespans')

        # create two allocations
        first = self.add_allocation(
            'timespans',
            datetime(2014, 2, 25, 12, 00),
            datetime(2014, 2, 25, 15, 00),
        )
        second = self.add_allocation(
            'timespans',
            datetime(2014, 2, 25, 16, 00),
            datetime(2014, 2, 25, 18, 00),
        )

        # reserve both in one go
        browser.open(self.infolder('/timespans/search'))

        browser.set_date('recurrence_start', datetime(2014, 2, 25))
        browser.set_date('recurrence_end', datetime(2014, 2, 25))

        browser.getControl(name='form.buttons.search').click()
        browser.getControl('Reserve selected').click()
        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        # while we're at it, make sure both manage views have the same dates
        manage_first = self.allocation_menu(*first)['manage']
        manage_second = self.allocation_menu(*second)['manage']

        browser.open(manage_first)
        self.assertIn('Feb 25, 2014 12:00 PM - 03:00 PM', browser.contents)
        self.assertIn('Feb 25, 2014 04:00 PM - 06:00 PM', browser.contents)

        browser.open(manage_second)
        self.assertIn('Feb 25, 2014 12:00 PM - 03:00 PM', browser.contents)
        self.assertIn('Feb 25, 2014 04:00 PM - 06:00 PM', browser.contents)

        # revoke the last date
        browser.getLink('Revoke', index=2).click()

        self.assertNotIn('12:00 PM - 03:00 PM', browser.contents)
        self.assertIn('04:00 PM - 06:00 PM', browser.contents)

        browser.getControl('Revoke').click()

        # ensure the date is gone
        browser.open(manage_first)
        self.assertIn('12:00 PM - 03:00 PM', browser.contents)
        self.assertNotIn('04:00 PM - 06:00 PM', browser.contents)

        # the second allocation now has no link to any reservations
        browser.open(manage_second)
        self.assertNotIn('12:00 PM - 03:00 PM', browser.contents)
        self.assertNotIn('04:00 PM - 06:00 PM', browser.contents)

        # remove the second date, ensuring that the whole reservation is gone
        browser.open(manage_first)
        browser.getLink('Revoke', index=1).click()
        self.assertIn('12:00 PM - 03:00 PM', browser.contents)

        browser.getControl('Revoke').click()

        browser.open(manage_first)
        self.assertNotIn('12:00 PM - 03:00 PM', browser.contents)
        self.assertNotIn('04:00 PM - 06:00 PM', browser.contents)

    @db.serialized
    def test_search_specific_time(self):
        # make sure that a search for a specific timerange on a partly
        # available allocation only returns the availability in this timerange
        browser = self.admin_browser

        self.add_resource('resource')

        daterange = (
            datetime(2014, 8, 20, 8, 00), datetime(2014, 8, 20, 10, 00)
        )

        allocation = ['resource', daterange[0], daterange[1]]
        self.add_allocation(*allocation, partly_available=True)

        # reserve 08:00 - 09:00
        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 21)
        )
        browser.getControl('Start time').value = '08:00 AM'
        browser.getControl('End time').value = '09:00 AM'
        browser.getControl('Available only').selected = True

        browser.getControl(name='form.buttons.search').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('100%', browser.contents)
        self.assertIn('event-available', browser.contents)

        browser.getControl('Reserve selected').click()

        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()

        # now search for 08:00 - 09:00, which should yield 0% (unavailable)
        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 21)
        )
        browser.getControl('Start time').value = '08:00 AM'
        browser.getControl('End time').value = '09:00 AM'
        browser.getControl('Available only').selected = False

        browser.getControl(name='form.buttons.search').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('0%', browser.contents)
        self.assertNotIn('100%', browser.contents)
        self.assertIn('event-unavailable', browser.contents)

        # now try to search for 09:00 - 10:00, which should yield 100%
        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 21)
        )
        browser.getControl('Start time').value = '09:00 AM'
        browser.getControl('End time').value = '10:00 AM'
        browser.getControl('Available only').selected = True

        browser.getControl(name='form.buttons.search').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('100%', browser.contents)
        self.assertIn('event-available', browser.contents)

    @db.serialized
    def test_search_and_reserve(self):
        browser = self.admin_browser

        self.add_resource('resource')

        dates = [
            (datetime(2014, 8, 20, 15, 00), datetime(2014, 8, 20, 16, 00)),
            (datetime(2014, 8, 21, 15, 00), datetime(2014, 8, 21, 16, 00)),
            (datetime(2014, 8, 22, 15, 00), datetime(2014, 8, 22, 16, 00)),
        ]

        allocations = [['resource', s, e] for s, e in dates]

        for allocation in allocations:
            self.add_allocation(*allocation)

        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 21)
        )

        # these values should be passed to the selection form
        browser.getControl('Start time').value = '10:00 AM'
        browser.getControl('End time').value = '10:00 PM'
        browser.getControl('Spots').value = '1'

        browser.getControl(name='form.buttons.search').click()
        self.assertNotIn('No results found', browser.contents)
        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)
        self.assertNotIn('Aug 22, 2014', browser.contents)

        # all allocations in the results are available, so they are preselected
        chks = browser.query('input[name="allocation_id"][checked="checked"]')
        self.assertEqual(len(chks), 2)

        browser.getControl('Reserve selected').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)
        self.assertNotIn('Aug 22, 2014', browser.contents)

        self.assertEqual(
            browser.query('#form-widgets-start_time').val(), '10:00 AM'
        )
        self.assertEqual(
            browser.query('#form-widgets-end_time').val(), '10:00 PM'
        )
        self.assertEqual(browser.query('#form-widgets-quota').val(), '1')

        # submit an invalid reservation and ensure that everything remains
        browser.getControl('Reserve').click()
        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)
        self.assertNotIn('Aug 22, 2014', browser.contents)

        self.assertEqual(
            browser.query('#form-widgets-start_time').val(), '10:00 AM'
        )
        self.assertEqual(
            browser.query('#form-widgets-end_time').val(), '10:00 PM'
        )
        self.assertEqual(browser.query('#form-widgets-quota').val(), '1')

        # go ahead
        browser.getControl('Email').value = 'info@example.org'
        browser.getControl('Reserve').click()

        # no try the search again, but include all dates
        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 22)
        )

        browser.getControl(name='form.buttons.search').click()
        self.assertNotIn('No results found', browser.contents)
        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)
        self.assertIn('Aug 22, 2014', browser.contents)

        # this time only one allocation is still available, which is reflected
        # in the state of the checkboxes
        chks = browser.query('input[name="allocation_id"][checked="checked"]')
        self.assertEqual(len(chks), 1)

    @db.serialized
    def test_change_time(self):
        browser = self.admin_browser

        self.add_resource('resource')

        allocations = [
            self.add_allocation(
                'resource',
                datetime(2014, 8, 20, 15, 00),
                datetime(2014, 8, 20, 16, 00),
                partly_available=True
            ),
            self.add_allocation(
                'resource',
                datetime(2014, 8, 21, 15, 00),
                datetime(2014, 8, 21, 16, 00),
                partly_available=False
            ),
        ]

        # reserve both allocations at once to see them in the same manage view
        browser.open(self.infolder('/resource/search'))
        browser.set_date(
            'form.widgets.recurrence_start', datetime(2014, 8, 20)
        )
        browser.set_date(
            'form.widgets.recurrence_end', datetime(2014, 8, 21)
        )
        browser.getControl(name='form.buttons.search').click()
        browser.getControl('Reserve selected').click()
        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        # two revoke links + one change link == 3
        browser.open(self.allocation_menu(*allocations[0])['manage'])
        self.assertEqual(browser.query('.timespan-actions a').length, 3)
        self.assertIn('Aug 20, 2014 03:00 PM - 04:00 PM', browser.contents)

        browser.getLink('Change').click()
        browser.getControl('Start').value = '03:30 PM'
        browser.getControl('Save').click()

        browser.open(self.allocation_menu(*allocations[0])['manage'])
        self.assertIn('Aug 20, 2014 03:30 PM - 04:00 PM', browser.contents)

    @db.serialized
    def test_remove_link_admin_only(self):
        browser = self.admin_browser

        self.add_resource('resource')

        self.add_allocation(
            'resource',
            datetime(2014, 8, 20, 15, 00),
            datetime(2014, 8, 20, 16, 00)
        ),
        self.add_allocation(
            'resource',
            datetime(2014, 8, 21, 15, 00),
            datetime(2014, 8, 21, 16, 00)
        )

        browser.open(self.infolder('/resource/search'))
        browser.set_date('recurrence_start', datetime(2014, 8, 20))
        browser.set_date('recurrence_end', datetime(2014, 8, 21))
        browser.getControl(name='form.buttons.search').click()
        self.assertIn('Delete selected', browser.contents)

        anonymous = self.new_browser()
        anonymous.open(self.infolder('/resource/search'))
        anonymous.set_date('recurrence_start', datetime(2014, 8, 20))
        anonymous.set_date('recurrence_end', datetime(2014, 8, 21))
        anonymous.getControl(name='form.buttons.search').click()
        self.assertNotIn('Delete selected', anonymous.contents)

    @db.serialized
    def test_remove_multiple_allocations(self):
        browser = self.admin_browser

        self.add_resource('resource')

        self.add_allocation(
            'resource',
            datetime(2014, 8, 20, 15, 00),
            datetime(2014, 8, 20, 16, 00)
        )
        self.add_allocation(
            'resource',
            datetime(2014, 8, 21, 15, 00),
            datetime(2014, 8, 21, 16, 00)
        )

        browser.open(self.infolder('/resource/search'))
        browser.set_date('recurrence_start', datetime(2014, 8, 20))
        browser.set_date('recurrence_end', datetime(2014, 8, 21))
        browser.getControl(name='form.buttons.search').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)

        browser.getLink('Delete selected').click()

        self.assertIn('Aug 20, 2014', browser.contents)
        self.assertIn('Aug 21, 2014', browser.contents)

        browser.getControl('Delete').click()

        browser.open(self.infolder('/resource/search'))
        browser.set_date('recurrence_start', datetime(2014, 8, 20))
        browser.set_date('recurrence_end', datetime(2014, 8, 21))
        browser.getControl(name='form.buttons.search').click()

        self.assertNotIn('Aug 20, 2014', browser.contents)
        self.assertNotIn('Aug 21, 2014', browser.contents)

    @db.serialized
    def test_remove_multiple_reserved_allocations(self):
        browser = self.admin_browser

        self.add_resource('resource')

        self.add_allocation(
            'resource',
            datetime(2014, 8, 20, 15, 00),
            datetime(2014, 8, 20, 16, 00),
            quota=2
        )
        self.add_allocation(
            'resource',
            datetime(2014, 8, 21, 15, 00),
            datetime(2014, 8, 21, 16, 00),
            quota=2
        )

        browser.open(self.infolder('/resource/search'))
        browser.set_date('recurrence_start', datetime(2014, 8, 20))
        browser.set_date('recurrence_end', datetime(2014, 8, 21))
        browser.getControl(name='form.buttons.search').click()

        browser.getControl('Reserve selected').click()
        browser.getControl('Email').value = 'test@example.org'
        browser.getControl('Reserve').click()
        browser.getControl('Submit Reservations').click()

        browser.open(self.infolder('/resource/search'))
        browser.set_date('recurrence_start', datetime(2014, 8, 20))
        browser.set_date('recurrence_end', datetime(2014, 8, 21))
        browser.getControl(name='form.buttons.search').click()

        browser.getLink('Delete selected').click()
        browser.getControl('Delete').click()

        self.assertIn('An existing reservation', browser.contents)



import mock
import pytz

from datetime import datetime, timedelta

from zope import i18n
from seantis.reservation.session import serialized
from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation.reports import GeneralReportParametersMixin
from seantis.reservation.reports.monthly_report import monthly_report
from seantis.reservation.reports.latest_reservations import (
    human_date,
    latest_reservations
)

reservation_email = u'test@example.com'


class TestReports(IntegrationTestCase):

    def test_report_parameters_mixin_defaults(self):
        self.login_admin()

        mixin = GeneralReportParametersMixin()

        mixin.request = self.request()
        mixin.context = self.create_resource()

        self.assertEqual(mixin.uuids, [mixin.context.uuid()])

        self.assertEqual(mixin.hidden_statuses, [])
        self.assertEqual(mixin.reservations, [])
        self.assertEqual(mixin.hidden_resources, [])
        self.assertEqual(mixin.show_details, False)

    def test_report_parameters_mixin_build_url(self):
        self.login_admin()

        mixin = GeneralReportParametersMixin()

        mixin.request = self.request()
        mixin.context = self.create_resource()

        mixin.request.set('hide_status', ['pending'])
        mixin.request.set('show_details', '1')
        mixin.request.set('hide_resource', 'test')

        extras = [('foo', 'bar')]

        expected = (
            'http://nohost/plone/seantis-reservation-resource/test?'
            'show_details=1&hide_status=pending&hide_resource=test&uuid={}'
            '&foo=bar'
        ).format(mixin.context.uuid())

        mixin.__name__ = 'test'  # build_url expects this, usually set by grok
        self.assertEqual(mixin.build_url(extras), expected)

    def test_monthly_report_empty(self):
        self.login_admin()

        resource = self.create_resource()
        report = monthly_report(2013, 9, {resource.uuid(): resource})

        self.assertEqual(len(report), 0)

    @serialized
    def test_monthly_report_reservations(self):
        self.login_admin()

        resource = self.create_resource()
        sc = resource.scheduler()

        today = (datetime(2013, 9, 25, 8), datetime(2013, 9, 25, 10))
        tomorrow = (datetime(2013, 9, 26, 8), datetime(2013, 9, 26, 10))

        sc.allocate(today, quota=2)
        sc.allocate(tomorrow, quota=1)

        sc.approve_reservations(sc.reserve(reservation_email, today))
        sc.approve_reservations(sc.reserve(reservation_email, today))
        sc.approve_reservations(sc.reserve(reservation_email, tomorrow))

        report = monthly_report(2013, 9, {resource.uuid(): resource})

        # one record for each day
        self.assertEqual(len(report), 2)

        # one resource for each day
        self.assertEqual(len(report[25]), 1)
        self.assertEqual(len(report[26]), 1)

        # two reservations on the first day
        self.assertEqual(len(report[25][resource.uuid()]['approved']), 2)

        # on reservation on the second day
        self.assertEqual(len(report[26][resource.uuid()]['approved']), 1)

    @mock.patch('seantis.reservation.utils.utcnow')
    def test_latest_reservations_human_date(self, utcnow):
        translate = lambda text: i18n.translate(
            text, target_language='en', domain='seantis.reservation'
        )
        human = lambda date: translate(human_date(date))

        utc = lambda *args: datetime(*args).replace(tzinfo=pytz.utc)

        utcnow.return_value = utc(2013, 9, 27, 11, 0)

        self.assertEqual(
            human(utc(2013, 9, 27, 21, 0)),
            u'Today, at 21:00'
        )

        self.assertEqual(
            human(utc(2013, 9, 27, 10, 0)),
            u'Today, at 10:00'
        )

        self.assertEqual(
            human(utc(2013, 9, 27, 0, 0)),
            u'Today, at 00:00'
        )

        self.assertEqual(
            human(utc(2013, 9, 26, 23, 59)),
            u'Yesterday, at 23:59'
        )

        self.assertEqual(
            human(utc(2013, 9, 26, 0, 0)),
            u'Yesterday, at 00:00'
        )

        self.assertEqual(
            human(utc(2013, 9, 25, 23, 59)),
            u'2 days ago, at 23:59'
        )

        self.assertEqual(
            human(utc(2013, 9, 25, 00, 00)),
            u'2 days ago, at 00:00'
        )

        self.assertEqual(
            human(utc(2013, 9, 24, 23, 59)),
            u'3 days ago, at 23:59'
        )

        # does not really deal with the future, it's not a concern
        self.assertEqual(
            human(utc(2014, 9, 27, 21, 0)),
            u'Today, at 21:00'
        )

    @serialized
    @mock.patch('seantis.reservation.utils.utcnow')
    def test_latest_reservations(self, utcnow):

        self.login_admin()

        resource = self.create_resource()
        sc = resource.scheduler()

        today = (datetime(2013, 9, 25, 8), datetime(2013, 9, 25, 10))
        tomorrow = (datetime(2013, 9, 26, 8), datetime(2013, 9, 26, 10))

        sc.allocate(today, quota=1)
        sc.allocate(tomorrow, quota=1)

        sc.approve_reservations(sc.reserve(reservation_email, today))
        sc.approve_reservations(sc.reserve(reservation_email, tomorrow))

        now = datetime.utcnow().replace(tzinfo=pytz.utc)

        daterange = (now - timedelta(days=30), now)

        report = latest_reservations({resource.uuid(): resource}, daterange)
        self.assertEqual(len(report), 2)

        daterange = (now - timedelta(days=31), now - timedelta(seconds=60))
        report = latest_reservations({resource.uuid(): resource}, daterange)
        self.assertEqual(len(report), 0)



import unittest2 as unittest

from sqlalchemy import create_engine

from zope import event
from zope.component import getUtility
from zope.security.management import newInteraction, endInteraction

from Acquisition import aq_base
from zope.component import getSiteManager
from Products.CMFPlone.tests.utils import MockMailHost
from Products.MailHost.interfaces import IMailHost

from plone import api
from plone.testing import z2
from plone.dexterity.utils import createContentInContainer
from plone.app.testing import TEST_USER_NAME, TEST_USER_ID
from plone.app.testing import login, logout, setRoles

from collective.betterbrowser import new_browser

from seantis.reservation import setuphandlers
from seantis.reservation.utils import getSite
from seantis.reservation.session import ISessionUtility

from seantis.reservation.testing import SQL_INTEGRATION_TESTING
from seantis.reservation.testing import SQL_FUNCTIONAL_TESTING

from seantis.reservation import maintenance

from Products.CMFCore.utils import getToolByName


class TestCase(unittest.TestCase):

    def setUp(self):

        # treat sqlalchemy warnings as errors
        import warnings
        from sqlalchemy.exc import SAWarning
        warnings.simplefilter("error", SAWarning)

        self.app = self.layer['app']
        self.portal = self.layer['portal']

        # setup mock mail host
        self._original_MailHost = self.portal.MailHost
        self.portal.MailHost = mailhost = MockMailHost('MailHost')
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(mailhost, provided=IMailHost)

        self.portal.email_from_address = 'noreply@example.com'

        # remove all test event subscribers
        event.subscribers = [
            e for e in event.subscribers if type(e) != TestEventSubscriber
        ]
        setuphandlers.dbsetup(None)

        self.setup_expected_date_formats()
        self.logged_in = False

    def setup_expected_date_formats(self):
        name_root = 'Products.CMFPlone.i18nl10n.override_dateformat.'
        api.portal.set_registry_record(
            name=name_root + 'Enabled',
            value=True
        )

        api.portal.set_registry_record(
            name=name_root + 'date_format_short',
            value='%d.%m.%Y'
        )

        api.portal.set_registry_record(
            name=name_root + 'date_format_long',
            value='%d.%m.%Y %H:%M'
        )

        api.portal.set_registry_record(
            name=name_root + 'time_format',
            value='%H:%M'
        )

    def tearDown(self):

        # reset original mail host
        self.portal.MailHost = self._original_MailHost
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(
            aq_base(self._original_MailHost), provided=IMailHost
        )

        util = getUtility(ISessionUtility)
        util.sessionstore.readonly.rollback()
        util.sessionstore.serial.rollback()

        maintenance.clear_clockservers()

        # since the testbrowser may create different records we need
        # to clear the database by hand each time
        outlaw = create_engine(util._dsn_cache['plone'])
        outlaw.execute('DELETE FROM reservations')
        outlaw.execute('DELETE FROM reserved_slots')
        outlaw.execute('DELETE FROM allocations')
        outlaw.dispose()

        self.logout()

        util._reset_sessions()

    def request(self):
        return self.layer['request']

    def login_manager(self):
        login(self.portal, TEST_USER_NAME)
        self.reset_test_user_roles()

    def set_test_user_roles(self, roles):
        setRoles(self.portal, TEST_USER_ID, roles)

    def reset_test_user_roles(self):
        self.set_test_user_roles(['Manager'])

    def login_admin(self):
        z2.login(self.app['acl_users'], 'admin')

    def new_browser(self):
        return new_browser(self.layer)

    def assign_reservation_manager(self, email, resource):
        username = email.split('@')[0]
        password = 'hunter2'

        acl_users = getToolByName(self.portal, 'acl_users')
        acl_users.userFolderAddUser(username, password, ['Member'], [])

        resource.manage_setLocalRoles(username, ['Reservation-Manager'])

        user = acl_users.getUser(username)
        properties = acl_users.mutable_properties.getPropertiesForUser(user)
        properties._properties['email'] = email
        acl_users.mutable_properties.setPropertiesForUser(user, properties)

        # the email will be there after the next retrieval of the user
        user = acl_users.getUser(username)
        assert user.getProperty('email') == email

        return username, password

    def logout(self):
        logout()

    def create_resource(self):
        return createContentInContainer(
            getSite(), 'seantis.reservation.resource'
        )

    def subscribe(self, eventclass):
        subscriber = TestEventSubscriber(eventclass)
        event.subscribers.append(subscriber)
        return subscriber

    @property
    def mailhost(self):
        return self.portal.MailHost


class TestEventSubscriber(object):

    def __init__(self, eventclass):
        self.eventclass = eventclass
        self.event = None

    def __call__(self, event):
        if type(event) is self.eventclass:
            self.event = event

    def was_fired(self):
        return self.event is not None

    def reset(self):
        self.event = None


# to use with integration where security interactions need to be done manually
class IntegrationTestCase(TestCase):
    layer = SQL_INTEGRATION_TESTING

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        # setup security
        newInteraction()

    def tearDown(self):
        endInteraction()
        super(IntegrationTestCase, self).tearDown()


# to use with the browser which does it's own security interactions
class FunctionalTestCase(TestCase):
    layer = SQL_FUNCTIONAL_TESTING



from logging import getLogger
log = getLogger('seantis.reservation')

from uuid import UUID
from uuid import uuid1 as new_uuid
from datetime import datetime, timedelta, MINYEAR, MAXYEAR
from itertools import groupby

from zope.event import notify

from sqlalchemy.sql import and_, or_, not_
from sqlalchemy.orm import joinedload, exc
from sqlalchemy import func, null

from seantis.reservation.events import (
    ReservationsMadeEvent,
    ReservationsApprovedEvent,
    ReservationsDeniedEvent,
    ReservationsRevokedEvent,
    ReservationsConfirmedEvent,
    ReservationTimeChangedEvent
)

from seantis.reservation.models import (
    Allocation, ReservedSlot, Reservation
)
from seantis.reservation.error import (
    AffectedPendingReservationError,
    AffectedReservationError,
    AlreadyReservedError,
    InvalidAllocationError,
    InvalidQuota,
    InvalidReservationError,
    InvalidReservationToken,
    NoReservationsToConfirm,
    NotReservableError,
    OverlappingAllocationError,
    QuotaImpossible,
    QuotaOverLimit,
    ReservationParametersInvalid,
    ReservationTooLong,
    TimerangeTooLong,
)

from seantis.plonetools.schemafields import validate_email

from seantis.reservation.session import serialized
from seantis.reservation.raster import rasterize_span, MIN_RASTER_VALUE
from seantis.reservation import utils
from seantis.reservation import Session


def all_allocations_in_range(start, end):
    # Query version of utils.overlaps
    return Session.query(Allocation).filter(
        or_(
            and_(
                Allocation._start <= start,
                start <= Allocation._end
            ),
            and_(
                start <= Allocation._start,
                Allocation._start <= end
            )
        )
    )


def availability_by_allocations(allocations):
    """Takes any iterator with alloctions and calculates the availability.
    Counts missing mirrors as 100% free and returns a value between 0-100 in
    any case.
    For single allocations check the allocation.availability property.

    """
    total, expected_count, count = 0, 0, 0
    for allocation in allocations:
        total += allocation.availability
        count += 1

        # Sum up the expected number of allocations. Missing allocations
        # indicate mirrors that have not yet been physically created.
        if allocation.is_master:
            expected_count += allocation.quota

    if not expected_count:
        return 0

    missing = expected_count - count
    total += missing * 100

    return total / expected_count


def availability_by_range(start, end, resources, is_exposed):
    """Returns the availability for the given resources in the given range.
    The callback *is_exposed* is used to check if the allocation is visible
    to the current user. This should usually be the exposure.for_allocations
    return value.

    """

    query = all_allocations_in_range(start, end)
    query = query.filter(Allocation.mirror_of.in_(resources))
    query = query.options(joinedload(Allocation.reserved_slots))

    allocations = (a for a in query if is_exposed(a))
    return availability_by_allocations(allocations)


def availability_by_day(start, end, resources, is_exposed):
    """Availability by range with a twist. Instead of returning a grand total,
    a dictionary is returned with each day in the range as key and a tuple of
    availability and the resources counted for that day.

    WARNING, this function should run as linearly as possible as a lot
    of records might be processed.

    """
    query = all_allocations_in_range(start, end)
    query = query.filter(Allocation.mirror_of.in_(resources))
    query = query.options(joinedload(Allocation.reserved_slots))
    query = query.order_by(Allocation._start)

    group = groupby(query, key=lambda a: a._start.date())
    days = {}

    for day, allocations in group:

        exposed = []
        members = set()

        for a in (a for a in allocations if is_exposed(a)):
            members.add(a.mirror_of)
            exposed.append(a)

        if not exposed:
            continue

        total = availability_by_allocations(exposed)

        days[day] = (total, members)

    return days


def reservations_by_session(session_id):

    # be sure to not query for all reservations. since a query should be
    # returned in any case we just use an impossible clause

    # this is mainly a security feature
    if not session_id:
        log.warn('Empty session id')
        return Session.query(Reservation).filter("0=1")

    query = Session.query(Reservation)
    query = query.filter(Reservation.session_id == session_id)
    query = query.order_by(Reservation.created)

    return query


@serialized
def confirm_reservations_for_session(session_id, token=None, language=None):
    """ Confirms all reservations of the given session id. Optionally
    confirms only the reservations with the given token. All if None.

    The langauge given is the language used for the confirmation event (which
    leads to an email sent in this language). Default is the site language.

    """

    assert session_id

    if not language:
        language = utils.get_current_site_language()

    reservations = reservations_by_session(session_id)

    if token:
        reservations = reservations.filter(Reservation.token == token)

    reservations = reservations.all()

    if not reservations:
        raise NoReservationsToConfirm

    for reservation in reservations:
        reservation.session_id = None

    notify(ReservationsConfirmedEvent(reservations, language))


@serialized
def remove_reservations_from_session(session_id, token):
    """ Removes the reservations with the given session_id and token. """

    assert token and session_id

    reservations = reservations_by_session(session_id)
    reservations = reservations.filter(Reservation.token == token)
    reservations = reservations.order_by(False)

    reservations.delete('fetch')

    # if we get here the token must be valid, we should then check if the
    # token is used in the reserved slots, because with autoapproval these
    # slots may be created straight away.

    slots = Session.query(ReservedSlot).filter(
        ReservedSlot.reservation_token == token
    )

    slots.delete('fetch')

    # we also update the timestamp of existing reservations within
    # the same session to ensure that we account for the user's activity
    # properly during the session expiration cronjob. Otherwise it is
    # possible that a user removes the latest reservations only to see
    # the rest of them vanish because his older reservations were
    # already old enough to be counted as expired.

    query = Session.query(Reservation)
    query = query.filter(Reservation.session_id == session_id)

    query.update({"modified": utils.utcnow()})


def find_expired_reservation_sessions(expiration_date):
    """ Goes through all reservations and returns the session ids of the
    unconfirmed ones which are older than the given expiration date.
    By default the expiration date is now - 15 minutes.

    Note that this method goes through ALL RESERVATIONS OF THE DATABASE. If
    this is not desired have a look at buildout/database.cfg.example to
    setup each site with its own database.

    """

    expiration_date = expiration_date or (
        utils.utcnow() - timedelta(minutes=15)
    )

    # first get the session ids which are expired
    query = Session.query(
        Reservation.session_id,
        func.max(Reservation.created),
        func.max(Reservation.modified)
    )

    query = query.group_by(Reservation.session_id)

    # != null() because != None is not allowed by PEP8
    query = query.filter(Reservation.session_id != null())

    # the idea is to remove all reservations belonging to sessions whose
    # latest update is expired - either delete the whole session or let
    # all of it be
    expired_sessions = []

    for session_id, created, modified in query.all():

        modified = modified or created
        assert created and modified

        if max(created, modified) < expiration_date:
            expired_sessions.append(session_id)

    return expired_sessions


@serialized
def remove_expired_reservation_sessions(expiration_date=None):
    """ Removes all reservations from all databases which have an
    expired session_id.

    Since this only concerns 'old' sessions it shouldn't be a problem
    however.

    """

    expired_sessions = find_expired_reservation_sessions(expiration_date)

    # remove those session ids
    if expired_sessions:
        reservations = Session.query(Reservation)
        reservations = reservations.filter(
            Reservation.session_id.in_(expired_sessions)
        )

        slots = Session.query(ReservedSlot)
        slots = slots.filter(
            ReservedSlot.reservation_token.in_(
                reservations.with_entities(Reservation.token).subquery()
            )
        )

        slots.delete('fetch')
        reservations.delete('fetch')

    return expired_sessions


@serialized
def extinguish_resource(resource_uuid):
    """ WARNING:
    Completely removes any trace of the given resource uuid. That means
    all reservations, reserved slots and allocations!

    """

    # set the language as this may be called when deleting plone which
    # will render all plone utils calls, like portal_languages, useless
    scheduler = Scheduler(resource_uuid, language='en')
    scheduler.managed_reservations().delete('fetch')
    scheduler.managed_reserved_slots().delete('fetch')
    scheduler.managed_allocations().delete('fetch')


class Scheduler(object):
    """Used to manage a resource as well as all connected mirrors.

    Master -> Mirror relationship
    =============================

    Mirrors are viewed as resources which mirror a master resource. These
    mirrors do not really exist as seantis.reservation.resource types in plone
    (unlike the master). Instead they have their own resource uuids which are
    calculated by creating new uuids from the master's uuid and the number of
    the mirror (see utils.generate_uuids).

    The reason for this mechanism is to ensure two things:

    - No more mirrors than required are created (if we tried that we would get
      integrity errors as the resource plus the start-time are unique)

    - The reservation slot does not need to carry any information about the
      mirror. It just references a resource uuid

    Since we do not want to calculate these mirror uuids all the time, since it
    is a somewhat expensive calculations and because it is a bit of hassle, we
    store the master uuid in the mirror_of field of each allocation record.

    """

    def __init__(self, resource_uuid, is_exposed=None, language=None):

        try:
            self.uuid = UUID(resource_uuid)
        except AttributeError:
            self.uuid = resource_uuid

        self.is_exposed = is_exposed or (lambda allocation: True)

        # the language is used for the events that are being sent
        # if no language is given, then the current site language is used
        if language:
            self.language = language
        else:
            self.language = utils.get_current_site_language()

    # using the managed_* functions as a starting point for queries is a good
    # idea because the querie's scope will always be correctly limited

    def managed_reservations(self):
        """ The reservations managed by this scheduler / resource. """
        query = Session.query(Reservation)
        query = query.filter(Reservation.resource == self.uuid)

        return query

    def managed_allocations(self):
        """ The allocations managed by this scheduler / resource. """
        query = Session.query(Allocation)
        query = query.filter(Allocation.mirror_of == self.uuid)

        return query

    def managed_reserved_slots(self):
        """ The reserved_slots managed by this scheduler / resource. """
        uuids = self.managed_allocations().with_entities(Allocation.resource)

        query = Session.query(ReservedSlot)
        query = query.filter(ReservedSlot.resource.in_(uuids))

        return query

    @serialized
    def allocate(self, dates, raster=15, quota=None, partly_available=False,
                 grouped=False, approve_manually=True,
                 reservation_quota_limit=0, whole_day=False
                 ):
        """Allocates a spot in the calendar.

        An allocation defines a timerange which can be reserved. No
        reservations can exist outside of existing allocations. In fact any
        reserved slot will link to an allocation.

        An allocation may be available as a whole (to reserve all or nothing).
        It may also be partly available which means reservations can be made
        for parts of the allocation.

        If an allocation is partly available a raster defines the granularity
        with which a reservation can be made (e.g. a raster of 15min will
        ensure that reservations are at least 15 minutes long and start either
        at :00, :15, :30 or :45)

        The reason for the raster is mainly to ensure that different
        reservations trying to reserve overlapping times need the same keys in
        the reserved_slots table, ensuring integrity at the database level.

        Allocations may have a quota, which determines how many times an
        allocation may be reserved. Quotas are enabled using a master-mirrors
        relationship.

        The master is the first allocation to be created. The mirrors copies of
        that allocation. See Scheduler.__doc__

        """
        dates = utils.pairs(dates)

        group = new_uuid()
        quota = quota or 1

        # if the allocation is not partly available the raster is set to lowest
        # possible raster value
        raster = partly_available and raster or MIN_RASTER_VALUE

        # the whole day option results in the dates being aligned to
        # the beginning of the day / end of it -> not timezone aware!
        if whole_day:
            for ix, (start, end) in enumerate(dates):
                dates[ix] = utils.align_range_to_day(start, end)

        # Ensure that the list of dates contains no overlaps inside
        for start, end in dates:
            if utils.count_overlaps(dates, start, end) > 1:
                raise InvalidAllocationError

        # Make sure that this span does not overlap another master
        for start, end in dates:
            start, end = rasterize_span(start, end, raster)

            existing = self.allocations_in_range(start, end).first()
            if existing:
                raise OverlappingAllocationError(start, end, existing)

        # Write the master allocations
        allocations = []
        for start, end in dates:
            allocation = Allocation(raster=raster)
            allocation.start = start
            allocation.end = end
            allocation.resource = self.uuid
            allocation.quota = quota
            allocation.mirror_of = self.uuid
            allocation.partly_available = partly_available
            allocation.approve_manually = approve_manually
            allocation.reservation_quota_limit = reservation_quota_limit

            if grouped:
                allocation.group = group
            else:
                allocation.group = new_uuid()

            allocations.append(allocation)

        Session.add_all(allocations)

        return allocations

    @serialized
    def change_quota(self, master, new_quota):
        """ Changes the quota of a master allocation.

        Fails if the quota is already exhausted.

        When the quota is decreased a reorganization of the mirrors is
        triggered. Reorganizing means eliminating gaps in the chain of mirrors
        that emerge when reservations are removed:

        Initial State:
        1   (master)    Free
        2   (mirror)    Free
        3   (mirror)    Free

        Reservations are made:
        1   (master)    Reserved
        2   (mirror)    Reserved
        3   (mirror)    Reserved

        A reservation is deleted:
        1   (master)    Reserved
        2   (mirror)    Free     <-- !!
        3   (mirror)    Reserved

        Reorganization is performed:
        1   (master)    Reserved
        2   (mirror)    Reserved <-- !!
        3   (mirror)    Free     <-- !!

        The quota is decreased:
        1   (master)    Reserved
        2   (mirror)    Reserved

        In other words, the reserved allocations are moved to the beginning,
        the free allocations moved at the end. This is done to ensure that
        the sequence of generated uuids for the mirrors always represent all
        possible keys.

        Without the reorganization we would see the following after
        decreasing the quota:

        The quota is decreased:
        1   (master)    Reserved
        3   (mirror)    Reserved

        This would make it impossible to calculate the mirror keys. Instead the
        existing keys would have to queried from the database.

        """

        assert new_quota > 0, "Quota must be greater than 0"

        if new_quota == master.quota:
            return

        if new_quota > master.quota:
            master.quota = new_quota
            return

        # Make sure that the quota can be decreased
        mirrors = self.allocation_mirrors_by_master(master)
        allocations = [master] + mirrors

        free_allocations = [a for a in allocations if a.is_available()]

        required = master.quota - new_quota
        if len(free_allocations) < required:
            raise AffectedReservationError(None)

        # get a map pointing from the existing uuid to the newly assigned uuid
        reordered = self.reordered_keylist(allocations, new_quota)

        # unused keys are the ones not present in the newly assignd uuid list
        unused = set(reordered.keys()) - set(reordered.values()) - set((None,))

        # get a map for resource_uuid -> allocation.id
        ids = dict(((a.resource, a.id) for a in allocations))

        for allocation in allocations:

            # change the quota for all allocations
            allocation.quota = new_quota

            # the value is None if the allocation is not mapped to a new uuid
            new_resource = reordered[allocation.resource]
            if not new_resource:
                continue

            # move all slots to the mapped allocation id
            new_id = ids[new_resource]

            for slot in allocation.reserved_slots:
                # build a query here as the manipulation of mapped objects in
                # combination with the delete query below seems a bit
                # unpredictable given the cascading of changes

                query = Session.query(ReservedSlot)
                query = query.filter(and_(
                    ReservedSlot.resource == slot.resource,
                    ReservedSlot.allocation_id == slot.allocation_id,
                    ReservedSlot.start == slot.start
                ))
                query.update(
                    {
                        ReservedSlot.resource: new_resource,
                        ReservedSlot.allocation_id: new_id
                    }
                )

        # get rid of the unused allocations (always preserving the master)
        if unused:
            query = Session.query(Allocation)
            query = query.filter(Allocation.resource.in_(unused))
            query = query.filter(Allocation.id != master.id)
            query = query.filter(Allocation._start == master._start)
            query.delete('fetch')

    def reordered_keylist(self, allocations, new_quota):
        """ Creates the map for the keylist reorganzation.

        Each key of the returned dictionary is a resource uuid pointing to the
        resource uuid it should be moved to. If the allocation should not be
        moved they key-value is None.

        """
        masters = [a for a in allocations if a.is_master]
        assert(len(masters) == 1)

        master = masters[0]
        allocations = dict(((a.resource, a) for a in allocations))

        # generate the keylist (the allocation resources may be unordered)
        keylist = [master.resource]
        keylist.extend(utils.generate_uuids(master.resource, master.quota))

        # prefill the map
        reordered = dict(((k, None) for k in keylist))

        # each free allocation increases the offset by which the next key
        # for a non-free allocation is acquired
        offset = 0
        for ix, key in enumerate(keylist):
            if allocations[key].is_available():
                offset += 1
            else:
                reordered[key] = keylist[ix - offset]

        return reordered

    def allocation_by_id(self, id):
        query = self.managed_allocations()
        query = query.filter(Allocation.mirror_of == self.uuid)
        query = query.filter(Allocation.id == id)
        return query.one()

    def allocations_by_ids(self, ids):
        query = self.managed_allocations()
        query = query.filter(Allocation.id.in_(ids))
        query = query.order_by(Allocation._start)
        return query

    def allocations_by_group(self, group, masters_only=True):
        return self.allocations_by_groups([group], masters_only=masters_only)

    def allocations_by_groups(self, groups, masters_only=True):
        query = self.managed_allocations()
        query = query.filter(Allocation.group.in_(groups))

        if masters_only:
            query = query.filter(Allocation.resource == self.uuid)

        return query

    def allocations_in_range(self, start, end, masters_only=True):
        query = self.managed_allocations()
        query = query.filter(
            or_(
                and_(
                    Allocation._start <= start,
                    start <= Allocation._end
                ),
                and_(
                    start <= Allocation._start,
                    Allocation._start <= end
                )
            )
        )

        if masters_only:
            query = query.filter(Allocation.resource == self.uuid)

        return query

    def allocations_by_reservation(self, token, id=None):
        """ Returns the allocations for the reservation if it was *approved*,
        pending reservations return nothing. If you need to get the allocation
        a pending reservation might be targeting, use _target_allocations
        in model.reservation.

        """

        # TODO -> this is much too joiny, it was easier when we assumed
        # that there would be one reservation per token, now that there
        # is more than one reservation per token we should denormalize
        # this a little by adding the reservation_id to the reserved slot

        groups = self.managed_reservations()
        groups = groups.with_entities(Reservation.target)
        groups = groups.filter(Reservation.token == token)

        if id is not None:
            groups = groups.filter(Reservation.id == id)

        allocations = self.managed_allocations()
        allocations = allocations.with_entities(Allocation.id)
        allocations = allocations.filter(Allocation.group.in_(
            groups.subquery()
        ))

        query = self.managed_allocations()
        query = query.join(ReservedSlot)
        query = query.filter(
            ReservedSlot.reservation_token == token
        )
        query = query.filter(
            ReservedSlot.allocation_id.in_(
                allocations.subquery()
            )
        )
        return query

    def allocation_by_date(self, start, end):
        query = self.allocations_in_range(start, end)
        return query.one()

    def allocation_mirrors_by_master(self, master):
        return [s for s in master.siblings() if not s.is_master]

    def dates_by_group(self, group):
        query = self.allocations_by_group(group)
        query = query.with_entities(Allocation._start, Allocation._end)

        return query.all()

    def dates_by_allocation_ids(self, ids, start_time=None, end_time=None):

        for allocation in self.allocations_by_ids(ids).all():

            s = start_time or allocation.display_start.time()
            e = end_time or allocation.display_end.time()

            s, e = allocation.limit_timespan(s, e)

            yield s, e - timedelta(microseconds=1)

    def manual_approval_required(self, ids):
        """ Returns True if any of the allocations require manual approval. """
        query = self.allocations_by_ids(ids)
        query = query.filter(Allocation.approve_manually == True)

        return query.first() and True or False

    def render_allocation(self, allocation):
        return self.is_exposed(allocation)

    def availability(self, start=None, end=None):
        """Goes through all allocations and sums up the availability."""

        if not (start and end):
            start = datetime(MINYEAR, 1, 1)
            end = datetime(MAXYEAR, 12, 31)

        return availability_by_range(start, end, [self.uuid], self.is_exposed)

    @serialized
    def move_allocation(
            self, master_id, new_start=None, new_end=None,
            group=None, new_quota=None, approve_manually=None,
            reservation_quota_limit=0, whole_day=None):

        assert master_id
        assert any([new_start and new_end, group, new_quota])

        # Find allocation
        master = self.allocation_by_id(master_id)
        mirrors = self.allocation_mirrors_by_master(master)
        changing = [master] + mirrors
        ids = [c.id for c in changing]

        assert(group or master.group)

        # Simulate the new allocation
        new_start = new_start or master.start
        new_end = new_end or master.end

        if whole_day:
            new_start, new_end = utils.align_range_to_day(new_start, new_end)

        new = Allocation(start=new_start, end=new_end, raster=master.raster)

        # Ensure that the new span does not overlap an existing one
        existing_allocations = self.allocations_in_range(new.start, new.end)

        for existing in existing_allocations:
            if existing.id not in ids:
                raise OverlappingAllocationError(new.start, new.end, existing)

        for change in changing:

            if change.partly_available:
                # confirmed reservations
                for reservation in change.reserved_slots:
                    if not new.contains(reservation.start, reservation.end):
                        raise AffectedReservationError(reservation)

                # pending reservations
                if change.is_master:  # (mirrors return the same values)
                    for pending in change.pending_reservations.with_entities(
                            Reservation.start, Reservation.end):
                        if not new.contains(*pending):
                            raise AffectedPendingReservationError(pending)

            else:
                # confirmed reservations
                if change.start != new.start or change.end != new.end:
                    if len(change.reserved_slots):
                        raise AffectedReservationError(
                            change.reserved_slots[0]
                        )

                    if change.is_master and \
                            change.pending_reservations.count():
                        raise AffectedPendingReservationError(
                            change.pending_reservations[0]
                        )

        # the following attributes must be equal over all group members
        # (this still allows to use move_allocation to remove an allocation
        #  from an existing group by specifiying the new group)
        for allocation in self.allocations_by_group(group or master.group):

            if approve_manually is not None:
                allocation.approve_manually = approve_manually

            if reservation_quota_limit is not None:
                allocation.reservation_quota_limit = reservation_quota_limit

            if new_quota is not None and allocation.is_master:
                self.change_quota(allocation, new_quota)

        for change in changing:
            change.start = new.start
            change.end = new.end
            change.group = group or master.group

    @serialized
    def remove_allocation(self, id=None, groups=None):
        if id:
            master = self.allocation_by_id(id)
            allocations = [master]
            allocations.extend(self.allocation_mirrors_by_master(master))
        elif groups:
            allocations = self.allocations_by_groups(
                groups, masters_only=False
            )
        else:
            raise NotImplementedError

        for allocation in allocations:
            assert allocation.mirror_of == self.uuid, """
                Trying to delete an allocation from a different resource than
                the scheduler and context. This is a serious error or
                someone trying to something funny with the POST parameters.
            """

            if len(allocation.reserved_slots) > 0:
                raise AffectedReservationError(allocation.reserved_slots[0])

            if allocation.pending_reservations.count():
                raise AffectedPendingReservationError(
                    allocation.pending_reservations[0]
                )

        for allocation in allocations:
            if not allocation.is_transient:
                Session.delete(allocation)

    @serialized
    def reserve(self, email, dates=None, group=None, data=None,
                session_id=None, quota=1):
        """ First step of the reservation.

        Seantis.reservation uses a two-step reservation process. The first
        step is reserving what is either an open spot or a place on the
        waiting list.

        The second step is to actually write out the reserved slots, which
        is done by approving an existing reservation.

        Most checks are done in the reserve functions. The approval step
        only fails if there's no open spot.

        This function returns a reservation token which can be used to
        approve the reservation in approve_reservation.

        """

        assert (dates or group) and not (dates and group)

        validate_email(email)

        if group:
            dates = self.dates_by_group(group)

        dates = utils.pairs(dates)

        # First, the request is checked for saneness. If any requested
        # date cannot be reserved the request as a whole fails.
        for start, end in dates:

            # are the parameters valid?
            if abs((end - start).days) >= 1:
                raise ReservationTooLong

            if start > end or (end - start).seconds < 5 * 60:
                raise ReservationParametersInvalid

            # can all allocations be reserved?
            for allocation in self.allocations_in_range(start, end):

                # start and end are not rasterized, so we need this check
                if not allocation.overlaps(start, end):
                    continue

                assert allocation.is_master

                # with manual approval the reservation ends up on the
                # waitinglist and does not yet need a spot
                if not allocation.approve_manually:
                    if not self.find_spot(allocation, start, end):
                        raise AlreadyReservedError

                    free = self.free_allocations_count(allocation, start, end)
                    if free < quota:
                        raise AlreadyReservedError

                if allocation.reservation_quota_limit > 0:
                    if allocation.reservation_quota_limit < quota:
                        raise QuotaOverLimit

                if allocation.quota < quota:
                    raise QuotaImpossible

                if quota < 1:
                    raise InvalidQuota

        # ok, we're good to go
        token = new_uuid()
        reservations = []

        # groups are reserved by group-identifier - so all members of a group
        # or none of them. As such there's no start / end date which is defined
        # implicitly by the allocation
        def new_reservations_by_group(group):
            if group:
                reservation = Reservation()
                reservation.token = token
                reservation.target = group
                reservation.status = u'pending'
                reservation.target_type = u'group'
                reservation.resource = self.uuid
                reservation.data = data
                reservation.session_id = session_id
                reservation.email = email.strip()
                reservation.quota = quota

                yield reservation

        # all other reservations are reserved by start/end date
        def new_reservations_by_dates(dates):
            already_reserved_groups = set()

            for start, end in dates:
                for allocation in self.allocations_in_range(start, end):
                    if allocation.group in already_reserved_groups:
                        continue

                    if not allocation.overlaps(start, end):
                        continue

                    # automatically reserve the whole group if the allocation
                    # is part of a group
                    if allocation.in_group:
                        already_reserved_groups.add(allocation.group)

                        # I really want to use 'yield from'. Python 3 ftw!
                        for r in new_reservations_by_group(allocation.group):
                            yield r
                    else:
                        reservation = Reservation()
                        reservation.token = token
                        reservation.start, reservation.end = rasterize_span(
                            start, end, allocation.raster
                        )
                        reservation.target = allocation.group
                        reservation.status = u'pending'
                        reservation.target_type = u'allocation'
                        reservation.resource = self.uuid
                        reservation.data = data
                        reservation.session_id = session_id
                        reservation.email = email.strip()
                        reservation.quota = quota

                        yield reservation

        # create the reservations
        if group:
            reservations = tuple(new_reservations_by_group(group))
        else:
            reservations = tuple(new_reservations_by_dates(dates))

        if not reservations:
            raise InvalidReservationError

        for reservation in reservations:
            Session.add(reservation)

        notify(ReservationsMadeEvent(reservations, self.language))

        return token

    @serialized
    def _approve_reservation_record(self, reservation):
        # write out the slots
        slots_to_reserve = []

        if reservation.target_type == u'group':
            dates = self.dates_by_group(reservation.target)
        else:
            dates = ((reservation.start, reservation.end),)

        # the reservation quota is simply implemented by multiplying the
        # dates which are approved

        dates = dates * reservation.quota

        for start, end in dates:

            for allocation in self.reservation_targets(start, end):

                for slot_start, slot_end in \
                        allocation.all_slots(start, end):
                    slot = ReservedSlot()
                    slot.start = slot_start
                    slot.end = slot_end
                    slot.resource = allocation.resource
                    slot.reservation_token = reservation.token

                    # the slots are written with the allocation
                    allocation.reserved_slots.append(slot)
                    slots_to_reserve.append(slot)

                # the allocation may be a fake one, in which case we
                # must make it realz yo
                if allocation.is_transient:
                    Session.add(allocation)

        reservation.status = u'approved'

        if not slots_to_reserve:
            raise NotReservableError

        return slots_to_reserve

    @serialized
    def approve_reservations(self, token):
        """ This function approves an existing reservation and writes the
        reserved slots accordingly.

        Returns a list with the reserved slots.

        """

        slots_to_reserve = []

        reservations = self.reservations_by_token(token).all()

        for reservation in reservations:
            slots_to_reserve.extend(
                self._approve_reservation_record(reservation)
            )

        notify(ReservationsApprovedEvent(reservations, self.language))

        return slots_to_reserve

    @serialized
    def deny_reservation(self, token):
        """ Denies a pending reservation, removing it from the records and
        sending an email to the reservee.

        """

        query = self.reservations_by_token(token)
        query = query.filter(Reservation.status == u'pending')

        reservations = query.all()

        query.delete()

        notify(ReservationsDeniedEvent(reservations, self.language))

    @serialized
    def revoke_reservation(self, token, reason, id=None, send_email=True):
        """ Revoke a reservation and inform the user of that."""

        reason = reason or u''

        # sending the email first is no problem as it won't work if
        # an exception triggers later in the request
        reservations = self.reservations_by_token(token, id).all()

        notify(ReservationsRevokedEvent(
            reservations, self.language, reason, send_email
        ))

        self.remove_reservation(token, id)

    @serialized
    def remove_reservation(self, token, id=None):
        """ Removes all reserved slots of the given reservation token.

        Note that removing a reservation does not let the reservee know that
        his reservation has been removed.

        If you want to let the reservee know what happened,
        use revoke_reservation.

        The id is optional. If given, only the reservation with the given
        token AND id is removed.

        """

        slots = self.reserved_slots_by_reservation(token, id)

        for slot in slots:
            Session.delete(slot)

        self.reservations_by_token(token, id).delete('fetch')

    @serialized
    def change_email(self, token, new_email):
        for reservation in self.reservations_by_token(token).all():
            reservation.email = new_email

    @serialized
    def change_reservation_data(self, token, data):

        for reservation in self.reservations_by_token(token).all():
            reservation.data = data

    @serialized
    def change_reservation_time_candidates(self, tokens=None):
        """ Returns the reservations that fullfill the restrictions
        imposed by change_reservation_time.

        Pass a list of reservaiton tokens to further limit the results.

        """

        query = self.managed_reservations()
        query = query.filter(Reservation.status == 'approved')
        query = query.filter(Reservation.target_type == 'allocation')

        groups = self.managed_allocations().with_entities(Allocation.group)
        groups = groups.filter(Allocation.partly_available == True)

        query = query.filter(Reservation.target.in_(groups.subquery()))

        if tokens:
            query = query.filter(Reservation.token.in_(tokens))

        return query

    @serialized
    def change_reservation_time(
        self, token, id, new_start, new_end, send_email=True, reason=None
    ):
        """ Allows to change the timespan of a reservation under certain
        conditions:

        - The new timespan must be reservable inside the existing allocation.
          (So you cannot use this method to reserve another allocation)
        - The referenced allocation must not be in a group.
        - The referenced allocation must be partly available.
        - The referenced reservation must be approved

        There is really only one usecase that this function works for:

        The user wants to change the timespan of a reservation in a meeting
        room kind of setup where you have lots of partly-available
        allocations.

        Returns True if a change was made.

        Just like revoke_reservation, this function raises an event which
        includes a send_email flag and a reason which may be used to inform
        the user of the changes to his reservation.

        see seantis.reservation.interfaces.IReservationTimeChangedEvent

        """

        # check for the reservation first as the allocation won't exist
        # if the reservation has not been approved yet
        assert new_start and new_end

        existing_reservation = self.reservations_by_token(token, id).one()

        assert existing_reservation.status == 'approved', """
            Reservation must be approved.
        """

        # if there's nothing to change, do not change
        if existing_reservation.start == new_start:
            if existing_reservation.end == new_end:
                return False
            if existing_reservation.end == new_end - timedelta(microseconds=1):
                return False

        # will return raise a MultipleResultsFound exception if this is a group
        allocation = self.allocations_by_reservation(token, id).one()

        assert allocation.partly_available, """
            Allocation must be partly available.
        """

        if not allocation.contains(new_start, new_end):
            raise TimerangeTooLong()

        reservation_arguments = dict(
            email=existing_reservation.email,
            dates=(new_start, new_end),
            data=existing_reservation.data,
            quota=existing_reservation.quota
        )

        old_start = existing_reservation.display_start
        old_end = existing_reservation.display_end

        Session.begin_nested()

        try:
            self.remove_reservation(token, id)

            new_token = self.reserve(**reservation_arguments)
            new_reservation = self.reservations_by_token(new_token).one()
            new_reservation.id = id
            new_reservation.token = token

            self._approve_reservation_record(new_reservation)

            notify(ReservationTimeChangedEvent(
                reservation=new_reservation,
                language=self.language,
                old_time=(old_start, old_end),
                new_time=(new_start, new_end),
                send_email=send_email,
                reason=reason or u'',
            ))
        except:
            Session.rollback()
            raise
        else:
            Session.commit()

        return True

    def find_spot(self, master_allocation, start, end):
        """ Returns the first free allocation spot amongst the master and the
        mirrors. Honors the quota set on the master and will only try the
        master if the quota is set to 1.

        If no spot can be found, None is returned.

        """
        master = master_allocation
        if master.is_available(start, end):
            return master

        if master.quota == 1:
            return None

        mirrors = self.allocation_mirrors_by_master(master)
        tries = master.quota - 1

        for mirror in mirrors:
            if mirror.is_available(start, end):
                return mirror

            if tries >= 1:
                tries -= 1
            else:
                return None

    def search_allocations(
        self, start, end,
        days=None,
        minspots=0,
        available_only=False,
        whole_day='any',
        groups='any',
        strict=False
    ):
        """ Search allocations using a number of options. The date is split
        into date/time. All allocations between start and end date within
        the given time (on each day) are included.

        For example, start=01.01.2012 12:00 end=31.01.2012 14:00 will include
        all allocaitons in January 2012 which OVERLAP the given times. So an
        allocation starting at 11:00 and ending at 12:00 will be included!

        WARNING allocations not matching the start/end date may be included
        if they belong to a group from which a member *is* included!

        If that behavior is not wanted set 'strict' to True
        or set 'include_groups' to 'no' (though you won't get any groups then).

        Allocations which are included in this way will return True in the
        following expression:

        getattr(allocation, 'is_extra_result', False)

        :start:
            Include allocations starting on or after this date.

        :end:
            Include allocations ending on or before this date.

        :days:
            List of days which should be considered, a subset of:
            (['mo', 'tu', 'we', 'th', 'fr', 'sa', 'su'])

            If left out, all days are included.

        :minspots:
            Minimum number of spots reservable.

        :available_only:
            If True, unavailable allocations are left out
            (0% availability). Default is False.

        :whole_day:
            May have one of the following values: 'yes', 'no', 'any'

            If yes, only whole_day allocations are returned.
            If no, whole_day allocations are filtered out.
            If any (default), all allocations are included.

            Any is the same as leaving the option out.

        :include_groups:
            'any' if all allocations should be included.
            'yes' if only group-allocations should be included.
            'no' if no group-allocations should be included.

            See allocation.in_group to see what constitutes a group

        :strict:
            Set to True if you don't want groups included as a whole if a
            groupmember is found. See comment above.
        """

        assert start
        assert end

        assert whole_day in ('yes', 'no', 'any')
        assert groups in ('yes', 'no', 'any')

        if days:
            days_map = {
                'mo': 0,
                'tu': 1,
                'we': 2,
                'th': 3,
                'fr': 4,
                'sa': 5,
                'su': 6
            }

            # get the day from the map - if impossible take the verbatim value
            # this allows for using strings or integers
            days = set(days_map.get(day, day) for day in days)

        query = self.allocations_in_range(start, end)
        query = query.order_by(Allocation._start)

        allocations = []

        known_groups = set()
        known_ids = set()

        for allocation in query.all():

            if not self.is_exposed(allocation):
                continue

            s = datetime.combine(allocation.display_start.date(), start.time())
            e = datetime.combine(allocation.display_end.date(), end.time())

            if not allocation.overlaps(s, e):
                continue

            if days:
                if allocation.display_start.weekday() not in days:
                    continue

            if whole_day != 'any':
                if whole_day == 'yes' and not allocation.whole_day:
                    continue
                if whole_day == 'no' and allocation.whole_day:
                    continue

            # minspots means that we don't show allocations which cannot
            # be reserved with the required spots in one reservation
            # so we can disregard all allocations with a lower quota limit.
            #
            # the spots are later checked again for actual availability, but
            # that is a heavier check, so it doesn't belong here.
            if minspots:
                if allocation.reservation_quota_limit > 0:
                    if allocation.reservation_quota_limit < minspots:
                        continue

            if available_only:
                if not self.find_spot(allocation, s, e):
                    continue

            if minspots:
                availability = self.availability(
                    allocation.start, allocation.end
                )

                if (minspots / float(allocation.quota) * 100.0) > availability:
                    continue

            # keep track of allocations in groups as those need to be added
            # to the result, even though they don't match the search
            in_group = (
                allocation.group in known_groups or allocation.in_group
            )

            if in_group:
                known_groups.add(allocation.group)
                known_ids.add(allocation.id)

            if groups != 'any':
                if groups == 'yes' and not in_group:
                    continue
                if groups == 'no' and in_group:
                    continue

            allocations.append(allocation)

        if not strict and groups != 'no' and known_ids and known_groups:
            query = self.managed_allocations()
            query = query.filter(not_(Allocation.id.in_(known_ids)))
            query = query.filter(Allocation.group.in_(known_groups))

            for allocation in query.all():
                allocation.is_extra_result = True
                allocations.append(allocation)

            allocations.sort(key=lambda a: a._start)

        return allocations

    def free_allocations_count(self, master_allocation, start, end):
        """ Returns the number of free allocations between master_allocation
        and it's mirrors. """

        free_allocations = 0

        if master_allocation.is_available(start, end):
            free_allocations += 1

        if master_allocation.quota == 1:
            return free_allocations

        for mirror in self.allocation_mirrors_by_master(master_allocation):
            if mirror.is_available(start, end):
                free_allocations += 1

        return free_allocations

    def reservation_targets(self, start, end):
        """ Returns a list of allocations that are free within start and end.
        These allocations may come from the master or any of the mirrors.

        """
        targets = []

        query = all_allocations_in_range(start, end)
        query = query.filter(Allocation.resource == self.uuid)

        for master_allocation in query:

            if not master_allocation.overlaps(start, end):
                continue  # may happen because start and end are not rasterized

            found = self.find_spot(master_allocation, start, end)

            if not found:
                raise AlreadyReservedError

            targets.append(found)

        return targets

    def reserved_slots_by_reservation(self, token, id=None):
        """ Returns all reserved slots of the given reservation.
        The id is optional and may be used only return the slots from a
        specific reservation matching token and id.
        """

        assert token

        query = self.managed_reserved_slots()
        query = query.filter(ReservedSlot.reservation_token == token)

        if id is None:
            return query
        else:
            ids = self.allocations_by_reservation(token, id)
            ids = ids.with_entities(Allocation.id)
            return query.filter(
                ReservedSlot.allocation_id.in_(ids.subquery())
            )

    def reservations_by_token(self, token, id=None):
        query = self.managed_reservations()
        query = query.filter(Reservation.token == token)

        if id:
            query = query.filter(Reservation.id == id)

        try:
            query.first()
        except exc.NoResultFound:
            raise InvalidReservationToken

        return query

    def reservations_by_group(self, group):
        tokens = self.managed_reservations().with_entities(Reservation.token)
        tokens = tokens.filter(Reservation.target == group)

        return self.managed_reservations().filter(
            Reservation.token.in_(
                tokens.subquery()
            )
        )

    def reservations_by_allocation(self, allocation_id):
        master = self.allocation_by_id(allocation_id)

        return self.reservations_by_group(master.group)



models = None


class OtherModels(object):
    """ Mixin class which allows for all models to access the other model
    classes without causing circular imports. """

    @property
    def models(self):
        global models
        if not models:
            from seantis.reservation import models as m_
            models = m_

        return models



from itertools import groupby
from seantis.reservation.models.reservation import BoundTimespan


class CombinedReservations(object):

    __slots__ = ['reservation', 'combined_timespans']

    def __init__(self, first_reservation, timespans):
        self.reservation = first_reservation
        self.combined_timespans = timespans

    def timespans(self):
        return [(t[0], t[1]) for t in self.combined_timespans]

    def bound_timespans(self):
        return self.combined_timespans

    def __getattr__(self, key):
        return getattr(self.reservation, key)


def combine_reservations(reservations):
    """ Takes a list of reservations, groups them by token and iterates
    through them, creating reservation record like objects which reference
    the first data found in the list of reservations and all the timespans
    combined.

    seantis.reservation assumes that a reservation with one token and many
    different reservations uses the same reservation data on all records
    linked by this token. This is less flexible, but a lot easier to
    explain and understand by the user.

    For example, take these reservations:

    id - token - data - start - end
    01 - 0xabc - x: 1 - 01.01 - 02.01
    02 - 0xabc - x: 1 - 02.01 - 03.01

    Fed into this class they iteration yields this result as
    a MinimalReservation:

    reservation {
        token     => '0xabc',
        data      => x: 1
        timespans => [(01.01, 02.01, 0xabc, 01), (02.01, 03.01, 0xabc, 02)]
    }

    It's really all about grouping reservations in a way that makes the
    result transparently usable by the reservation list view.
    """

    by_token = lambda r: r.token
    reservations = sorted(reservations, key=by_token)

    for token, reservations in groupby(reservations, key=by_token):
        reservations = tuple(r for r in reservations)

        timespans = []
        for reservation in reservations:
            for start, end in reservation.timespans():
                timespans.append(
                    BoundTimespan(
                        start, end, reservation.token, reservation.id
                    )
                )

        yield CombinedReservations(
            first_reservation=reservations[0], timespans=timespans
        )



import re
import time
import json
import collections
import functools
import isodate
import base64
import sys

from copy import deepcopy
from datetime import datetime, timedelta, date, time as datetime_time
from urlparse import urljoin
from urllib import quote
from itertools import tee, izip
from uuid import UUID
from uuid import uuid5 as new_uuid_mirror
from plone.uuid.interfaces import IUUID, IUUIDAware
from plone import api

import pytz

from plone.dexterity.utils import SchemaNameEncoder

from App.config import getConfiguration
from Acquisition import aq_inner
from zope.component import getMultiAdapter
from zope.component.hooks import getSite
from zope import i18n
from zope import interface
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.i18nl10n import weekdayname_msgid_abbr, monthname_msgid
from z3c.form.interfaces import ActionExecutionError
from plone.i18n.locales.languages import _languagelist
from plone.app.textfield.value import RichTextValue

from OFS.interfaces import IApplication
from Products.CMFPlone.interfaces import IPloneSiteRoot

from seantis.reservation import error
from seantis.reservation import _


# avoid circular import of settings
_settings = None


def settings():
    global _settings

    if _settings is None:
        from seantis.reservation import settings
        _settings = settings

    return _settings


try:
    from collections import OrderedDict  # python >= 2.7
    OrderedDict  # Pyflakes
except ImportError:
    from ordereddict import OrderedDict  # python < 2.7
    OrderedDict  # Pyflakes

dexterity_encoder = SchemaNameEncoder()


class ConfigurationError(Exception):
    pass


def maybe_call(value):
    return value() if callable(value) else value


def profile(fn):
    """ Naive profiling of a function.. on unix systems only. """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        start = time.time()

        result = fn(*args, **kwargs)
        print fn.__name__, 'took', (time.time() - start) * 1000, 'ms'

        return result

    return wrapper


def additional_data_dictionary(data, fti):
    """ Takes the data from a post request and puts the relevant bits into
    a dictionary with the following structure:

    {
        "form_key": {
            "desc": "Form Description",
            "interface": "The interface used",
            "values": [
                {
                    "key": "value_key",
                    "sortkey": 1-n,
                    "value": "value",
                    "desc": "Value Description"
                },
                ...
            ]
        }
    }

    Interfaces used in the reservation data are the interfaces with the
    IReservationFormSet marker set.

    The dictionary is later converted to JSON and stored on the reservation.
    """

    result = dict()
    used_data = dict([(k, v) for k, v in data.items() if v is not None])

    def values(iface, ifacekey):
        for key, value in used_data.items():
            if not key.startswith(ifacekey + '.'):
                continue

            subkey = key.split('.')[1]
            desc = iface.getDescriptionFor(subkey).title
            sortkey = iface.get(subkey).order

            yield dict(key=subkey, desc=desc, value=value, sortkey=sortkey)

    for key, info in fti.items():
        desc, iface = info[0], info[1]

        record = dict()
        record['desc'] = desc
        record['interface'] = iface.getName()
        record['values'] = list(values(iface, key))

        if not record['values']:
            continue

        result[key] = record

    return result


def merge_data_dictionaries(base, extra):
    """ Merges the given data dictionaries. The extra dictionary will overwrite
    matching keys in the base dictionary.

    """

    merged = deepcopy(base)

    for extra_form in extra:
        if extra_form in base:
            merged[extra_form]['desc'] = extra[extra_form]['desc']
            merged[extra_form]['interface'] = extra[extra_form]['interface']

            # this really should be a dict, since the key matters...
            base_values = dict(
                (i['key'], i) for i in base[extra_form]['values']
            )
            extra_values = dict(
                (i['key'], i) for i in extra[extra_form]['values']
            )

            merged[extra_form]['values'] = []

            for value in base[extra_form]['values']:
                if value['key'] in extra_values:
                    merged[extra_form]['values'].append(
                        extra_values[value['key']]
                    )
                else:
                    merged[extra_form]['values'].append(
                        base_values[value['key']]
                    )

            for value in extra[extra_form]['values']:
                if value['key'] in base_values:
                    continue
                else:
                    merged[extra_form]['values'].append(value)
        else:
            merged[extra_form] = extra[extra_form]

    return merged


def additional_data_objects(data):
    """ Takes the additional data dictionary and returns a new dictionary
    with the keys being the formsets, and the values being objects.

    An 'address' formset with 'street' and 'town' will produce this:

    {'address': obj}

    Where obj can be acccessed like this:

    obj.street
    obj.town

    """
    result = {}

    class DynamicObject(object):
        pass

    for formset in data:
        obj = DynamicObject()
        for item in data[formset]['values']:
            setattr(obj, item['key'], item['value'])

        result[formset] = obj

    return result


def zope_root():
    this = getSite()

    while not IApplication.providedBy(this):
        this = this.aq_inner.aq_parent

    return this


def plone_sites(root=None):
    root = root or zope_root()

    sites = []
    for id, item in root.items():
        if not IPloneSiteRoot.providedBy(item):
            continue

        sites.append(item)
        sites.extend(plone_sites(item))

    return sites


def mock_data_dictionary(data, formset_key='mock', formset_desc='Mocktest'):
    """ Given a dictionary of key values this function returns a dictionary
    with the same structure as additional_data_dictionary, without the use
    of interfaces.

    This function exists for testing purposes.
    """

    records = list()

    formset = dict()
    formset['desc'] = formset_desc
    formset['interface'] = 'mockinterface'
    formset['values'] = records

    for i, item in enumerate(data.items()):
        records.append(
            dict(key=item[0], sortkey=i, value=item[1], desc=item[0])
        )

    return {formset_key: formset}


def count_overlaps(dates, start, end):
    count = 0

    for otherstart, otherend in dates:
        count += overlaps(start, end, otherstart, otherend) and 1 or 0

    return count


def overlaps(start, end, otherstart, otherend):
    if otherstart <= start and start <= otherend:
        return True

    if start <= otherstart and otherstart <= end:
        return True

    return False

_requestid_expr = re.compile(r'\d')


def request_id_as_int(string):
    """Returns the id of a request as int without throwing an error if invalid
    characters are in the requested string (like ?id=11.11).

    """
    if string is None:
        return 0

    return int(''.join(re.findall(_requestid_expr, string)))


class memoize(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


@memoize
def _exposure():
    # keep direct imports out of utils as it will lead to circular imports
    from seantis.reservation import exposure
    return exposure


def compare_link(resources):
    """Builds the compare link for the given list of resources. """

    # limit the resources already on the link, as the view needs
    # at least two visible resources to be even callable
    # (another check for security is made in the resources module later)
    resources = _exposure().limit_resources(resources)

    if len(resources) < 2:
        return ''

    link = resources[0:1][0].absolute_url_path() + '?'
    compare_to = [string_uuid(r) for r in resources[1:]]

    for uuid in compare_to:
        link += 'compare_to=' + string_uuid(uuid) + '&'

    return link.rstrip('&')


def export_link(context, request, resources):
    """Builds the reservations excel export link for the
    given list of resources.

    """

    if not resources:
        return ''

    if not _exposure().for_views(context, request)('reservation-exports'):
        return ''

    return ''.join((
        context.absolute_url(),
        '/reservation-exports?uuid=',
        '&uuid='.join(
            map(string_uuid, resources)
        )
    ))


def monthly_report_link(context, request, resources):
    """Builds the monthly report link given the list of the resources
    using the current year and month."""

    if not resources:
        return ''

    # ensure that the view may be called
    if not _exposure().for_views(context, request)('monthly_report'):
        return ''

    today = datetime.now()

    url = context.absolute_url()

    # note that 'monthly_report' is copied in reports.py
    url += '/monthly_report'
    url += '?year=' + str(today.year)
    url += '&month=' + str(today.month)

    for resource in resources:
        url += '&uuid=' + string_uuid(resource)

    return url


def latest_reservations_link(context, request, resources):
    """Builds the latest reservations report link. """
    if not resources:
        return ''

    # ensure that the view may be called
    if not _exposure().for_views(context, request)('latest_reservations'):
        return ''

    return '{}/{}?uuid={}'.format(
        context.absolute_url(), 'latest_reservations',
        '&uuid='.join(map(string_uuid, resources))
    )


def get_current_language(context, request):
    """Returns the current language of the request"""
    context = aq_inner(context)
    portal_state = getMultiAdapter(
        (context, request), name=u'plone_portal_state'
    )
    return portal_state.language()


def get_current_site_language():
    """Returns the current language of the current site"""

    context = getSite()
    portal_languages = getToolByName(context, 'portal_languages')
    langs = portal_languages.getSupportedLanguages()

    return langs and langs[0] or u'en'


def get_site_email_sender():
    """Returns the default sender of emails of the current site in the
    following format: 'sendername<sender@domain>'.

    """

    site = getSite()
    address = site.getProperty('email_from_address')
    name = site.getProperty('email_from_name')

    if not address:
        return None

    if name:
        return '%s<%s>' % (name, address)
    else:
        return address


def translator(context, request):
    """Returns a function which takes a single string and translates it using
    the curried values for context & request.

    """
    def curried(text):
        return translate(context, request, text)
    return curried


def translate(context, request, text, domain=None):
    """Translates the given text using context & request."""

    # xx-xx languages will not work here, though they work when Plone does
    # it in a template. For now it does not matter as we have no country
    # specific translation available
    lang = get_current_language(context, request).split('-')[0]
    return i18n.translate(text, target_language=lang, domain=domain)


def translate_workflow(context, request, text):
    return translate(context, request, text, domain='plone')


def native_language_name(code):
    return _languagelist[code]['native']


def handle_action(action=None, success=None, message_handler=None):
    """Calls 'action' and then 'success' if everything went well or
    'message_handler' with a message if an exception happened.

    If no message_handler is given, an ActionExecutionError is raised
    with the message attached.

    """
    try:
        result = None
        if action:
            result = action()
        if success:
            success()

        return result

    except Exception, e:
        e = hasattr(e, 'orig') and e.orig or e
        handle_exception(e, sys.exc_info(), message_handler)


def form_error(msg):
    raise ActionExecutionError(interface.Invalid(msg))


def handle_exception(ex, exception_info, message_handler=None):
    if isinstance(ex, error.CustomReservationError):
        form_error(ex.msg)

    msg = error.errormap.get(type(ex))

    if not msg:
        raise exception_info[1], None, exception_info[2]

    if message_handler:
        message_handler(msg)
    else:
        form_error(msg)

_uuid_regex = re.compile(
    '[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}'
)


def is_uuid(obj):
    """Returns true if the given obj is a uuid. The obj may be a string
    or of type UUID. If it's a string, the uuid is checked with a regex.
    """
    if isinstance(obj, basestring):
        return re.match(_uuid_regex, unicode(obj))

    return isinstance(obj, UUID)


def string_uuid(obj):
    """ Returns the uuid as string without hyphens. Takes either UUIDs, strings
    with hyphens, objects which are IUUIDAware or objects which have a UID
    attribute. (The latter works for catalog brains). """

    if isinstance(obj, basestring):
        obj = str(obj)
    elif hasattr(obj, 'UID'):
        if callable(obj.UID):
            obj = obj.UID()  # plone 4.3, because
        else:
            obj = obj.UID
    elif IUUIDAware.providedBy(obj):
        obj = IUUID(obj)
    elif callable(obj):
        obj = obj()

    return UUID(str(obj)).hex


def real_uuid(obj):
    """ Same as string_uuid but casted to uuid.UUID. """
    return UUID(string_uuid(obj))


# TODO cache this incrementally
def generate_uuids(uuid, quota):
    mirror = lambda n: new_uuid_mirror(uuid, str(n))
    return [mirror(n) for n in xrange(1, quota)]


def uuid_query(uuid):
    """ Returns a tuple of uuids for querying the zodb. See why:
    http://stackoverflow.com/questions/10137632/
    querying-portal-catalog-using-typed-uuids-instead-of-string-uuids

    """

    uuid = string_uuid(uuid)

    if '-' in uuid:
        return (uuid.replace('-', ''), uuid)
    return (uuid, '-'.join([
        uuid[:8], uuid[8:12], uuid[12:16], uuid[16:20], uuid[20:]]))


def get_resource_by_uuid(
    uuid, ensure_portal_type='seantis.reservation.resource'
):
    """Returns the zodb object with the given uuid."""
    catalog = getToolByName(getSite(), 'portal_catalog')

    if ensure_portal_type:
        results = catalog(UID=uuid_query(uuid), portal_type=ensure_portal_type)
    else:
        results = catalog(UID=uuid_query(uuid))

    return len(results) == 1 and results[0] or None


def get_resource_title(resource, title_prefix=''):
    if hasattr(resource, '__parent__'):
        parent = resource.__parent__.title
    elif hasattr(resource, 'parent'):
        parent = resource.parent().title
    else:
        return title_prefix + resource.title

    return ' - '.join((parent, title_prefix + resource.title))


def get_reservation_quota_statement(quota):
    if quota > 1:
        return _(u'<b>${quota}</b> reservations', mapping={'quota': quota})
    else:
        return _(u'<b>1</b> reservation')


class UUIDEncoder(json.JSONEncoder):
    """Encodes UUID objects as string in JSON."""
    def default(self, obj):
        if isinstance(obj, UUID):
            return string_uuid(obj)
        return json.JSONEncoder.default(self, obj)


class UserFormDataEncoder(json.JSONEncoder):
    """Encodes additional user data."""

    def default(self, obj):

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, datetime):
            return u'__datetime__@%s' % isodate.datetime_isoformat(obj)

        if isinstance(obj, date):
            return u'__date__@%s' % isodate.date_isoformat(obj)

        if isinstance(obj, datetime_time):
            return u'__time__@%s' % isodate.time_isoformat(obj)

        if isinstance(obj, RichTextValue):
            return u'__richtext__@%s' % base64.b64encode(json.dumps(dict(
                raw=obj.raw,
                encoding=obj.encoding,
                mime=obj.mimeType,
                output_mime=obj.outputMimeType
            )))

        return json.JSONEncoder.default(self, obj)


def userformdata_decode(string):
    if not isinstance(string, basestring):
        return string

    if string.startswith(u'__date__@'):
        return isodate.parse_date(string[9:19])

    if string.startswith(u'__datetime__@'):
        return isodate.parse_datetime(string[13:32])

    if string.startswith(u'__time__@'):
        return isodate.parse_time(string[9:18])

    if string.startswith(u'__richtext__@'):
        data = json.loads(base64.b64decode(string[13:]))
        return RichTextValue(
            raw=data['raw'],
            mimeType=data['mime'],
            outputMimeType=data['output_mime'],
            encoding=data['encoding']
        )

    return string


def as_human_readable_string(value):
    if isinstance(value, basestring):
        return value

    if isinstance(value, datetime):
        # don't use strftime here because users may end up entering year '3'
        # strftime does not suppport years before 1900, which is just lame
        # in this case we just use the German locale for now..
        if value.year < 1900:
            return '%02d.%02d.%04d %02d:%02d' % (
                value.day,
                value.month,
                value.year,
                value.hour,
                value.minute
            )
        else:
            return api.portal.get_localized_time(value, long_format=True)

    if isinstance(value, date):
        if value.year < 1900:
            return '%02d.%02d.%04d' % (
                value.day,
                value.month,
                value.year
            )
        else:
            dt = datetime(value.year, value.month, value.day)
            return api.portal.get_localized_time(dt, long_format=False)

    if isinstance(value, RichTextValue):
        return value.output

    if value is True:
        return _(u'Yes')

    if value is False:
        return _(u'No')

    if value is None:
        return u''

    if isinstance(value, (list, tuple)):
        return ', '.join(as_human_readable_string(v) for v in value)

    return value


def event_class(availability):
    """Returns the event class to be used depending on the availability."""

    s = settings()

    available = s.get('available_threshold')
    partly = s.get('partly_available_threshold')

    if availability >= available:
        return 'event-available'
    elif partly <= availability and availability < available:
        return 'event-partly-available'
    else:
        return 'event-unavailable'


def allocation_class(allocation):
    """Returns the class set on the fullcalendar event for the allocation."""
    if allocation.approve_manually:
        return 'approve-manually'
    else:
        return 'approve-automatically'


def event_availability(
    context, request, scheduler, allocation, start=None, end=None
):
    """ Returns the availability, the text with the availability and the class
    for the availability to display on the calendar view.

    If start and end are given and the allocation is partly_available
    the availability is set to 100% if the scheduler's find_spot method
    returns True. That is if the timespan between start and end
    is completely reservable.

    This is very involved and slow, so you probably shouldn't use that.
    This feature tries to account for the fact that parts of allocations
    can be reserved in search, where the user gets the impression that
    the time he entered is the actual allocation, when that timespan might
    only refer to a part of the allocation.

    For now this will be a new features which we'll test against. In the
    future this needs to be made much faster => TODO.

    """
    translate = translator(context, request)

    if start and end and allocation.partly_available:
        availability = scheduler.find_spot(allocation, start, end) and 100 or 0
    else:
        availability = scheduler.availability(allocation.start, allocation.end)

    spots = int(round(allocation.quota * availability / 100))

    # get the title shown on the calendar block
    if allocation.partly_available:
        text = translate(_(u'%i%% Free')) % availability
    else:
        if allocation.quota > 1:
            text = translate(_(u'%i/%i Spots Available')) % (
                spots, allocation.quota
            )
        else:
            text = translate(_(u'%i/%i Spot Available')) % (
                spots, allocation.quota
            )

    # with approval the number of people in the waitinglist have to be shown
    if allocation.approve_manually:
        length = allocation.waitinglist_length
        if length == 0:
            text += '\n' + translate(_(u'Waitinglist is Free'))
        elif length == 1:
            text += '\n' + translate(_(u'One Person Waiting'))
        else:
            text += '\n' + translate(_(u'%i People Waiting')) % length

    return (
        availability,
        text,
        ' '.join((event_class(availability), allocation_class(allocation)))
    )


def flatten(l):
    """Generator for flattening irregularly nested lists. 'Borrowed' from here:

    http://stackoverflow.com/questions/2158395/
    flatten-an-irregular-list-of-lists-in-python
    """
    for el in l:
        if isinstance(el, collections.Iterable) and \
                not isinstance(el, basestring):
            for sub in flatten(el):
                yield sub
        else:
            yield el


def pack(v):
    """Puts v in a list if not already an iterator."""
    return [v] if not hasattr(v, '__iter__') else v


def pairs(l):
    """Takes any list and returns pairs:
    ((a,b),(c,d)) => ((a,b),(c,d))
    (a,b,c,d) => ((a,b),(c,d))

    http://opensourcehacker.com/2011/02/23/
    tuplifying-a-list-or-pairs-in-python/
    """
    l = list(flatten(l))
    return zip(*[l[x::2] for x in (0, 1)])


def pairwise(iterable):
    """Almost like pairs, but not quite:
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    """
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)


def safe_parse_int(value, default):
    try:
        return int(value)
    except (ValueError, TypeError):
        return default


def get_config(key):
    config = getConfiguration()
    if not hasattr(config, 'product_config'):
        raise ConfigurationError('No configuration found.')

    configuration = config.product_config.get('seantis.reservation', dict())
    return configuration.get(key)


# obsolete in python 2.7
def total_timedelta_seconds(td):
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / \
        10 ** 6


def utc_mktime(utc_tuple):
    """Returns number of seconds elapsed since epoch

    Note that no timezone are taken into consideration.

    utc tuple must be: (year, month, day, hour, minute, second)

    """

    if len(utc_tuple) == 6:
        utc_tuple += (0, 0, 0)
    return time.mktime(utc_tuple) - time.mktime((1970, 1, 1, 0, 0, 0, 0, 0, 0))


def utctimestamp(dt):
    dt = dt.replace(tzinfo=pytz.utc)
    return utc_mktime(dt.timetuple())


def utcnow():
    """Returns the utc now function result with the correct timezone set. """
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def merge_reserved_slots(slots):
    """Given a list of reserved slots a list of tuples with from-to datetimes
    is formed, with adjacent slots being combined into one continious timespan.
    Usually this leaves reserved_slots be, but if the slots in the list come
    from a partially available allocation the reserved_slots of such an
    allocation can be merged into one timespan.

    So instead of
    08:00 - 08:14:59:59
    08:15 - 08:29:59:59

    You'll get
    08:00 - 08:30

    for this, the display_start and display_end of ResourceSlot are used."""

    class Timespan(object):
        def __init__(self, start, end):
            self.start = start
            self.end = end

    if len(slots) == 1:
        return [Timespan(slots[0].start, slots[0].end)]

    slots = sorted(slots, key=lambda s: s.start)

    merged = []
    current = Timespan(None, None)

    for this, next in pairwise(slots):

        if abs((next.start - this.end).seconds) <= 1:
            current.start = current.start and current.start or this.start
            current.end = next.end
        else:
            merged.append(current)
            current = Timespan(None, None)

    if current.start:
        merged.append(current)

    return merged


def urlparam(base, url, params):
    """Joins an url, adding parameters as query parameters."""
    if not base.endswith('/'):
        base += '/'

    urlquote = lambda fragment: quote(unicode(fragment).encode('utf-8'))
    querypair = lambda pair: pair[0] + '=' + urlquote(pair[1])

    query = '?' + '&'.join(map(querypair, params.items()))
    return ''.join(reduce(urljoin, (base, url, query)))


class EventUrls(object):
    def __init__(self, resource, request, exposure):
        self.resource = resource
        self.base = resource.absolute_url_path()
        self.request = request
        self.translate = translator(resource, request)
        self.menu = {}
        self.order = []
        self.exposure = exposure
        self.default = ""
        self.move = ""

    @memoize
    def restricted_url(self, view):
        """Returns a function which can be used to build an url with optional
        parameters. The function only builds the url if the current user has
        the right to do so.

        """
        base = self.resource.absolute_url_path()
        is_exposed = self.exposure.for_views(self.resource, self.request)

        def build(params):
            if is_exposed(view):
                return urlparam(base, view, params)
            else:
                return None

        # return closure
        return build

    @memoize
    def get_restricted(self, view, params):
        urlfactory = self.restricted_url(view)
        if not urlfactory:
            return None

        return urlfactory(params)

    def menu_add(self, group, name, view, params, target):
        url = self.get_restricted(view, params)
        if not url:
            return

        group = self.translate(group)
        name = self.translate(name)

        if not group in self.menu:
            self.menu[group] = []
            self.order.append(group)

        self.menu[group].append(dict(name=name, url=url, target=target))

    def default_url(self, view, params):
        url = self.get_restricted(view, params)
        if not url:
            return

        self.default = url

    def move_url(self, view, params):
        url = self.get_restricted(view, params)
        if not url:
            return

        self.move = url


def get_date_range(day, start_time, end_time):
    """Returns the date-range of a date a start and an end time."""
    start = datetime.combine(day, start_time)
    end = datetime.combine(day, end_time)

    # since the user can only one date with separate times it is assumed
    # that an end before a start is meant for the following day
    if end < start:
        end += timedelta(days=1)

    return start, end


def flash(context, message, type='info'):
    context.plone_utils.addPortalMessage(message, type)


def month_name(context, request, month):
    """Returns the text for the given month (1-12)."""
    assert (1 <= month and month <= 12)
    msgid = monthname_msgid(month)
    return translate(context, request, msgid, 'plonelocales')


def weekdayname_abbr(context, request, day):
    """Returns the text for the given day (0-6), 0 being sunday.
    Since 0 is sunday in the plone function and monday in the python stdlib
    you might want to shift the day.weekday() result"""
    assert (0 <= day and day <= 6)
    msgid = weekdayname_msgid_abbr(day)
    return translate(context, request, msgid, 'plonelocales')


def shift_day(stdday):
    return abs(0 - (stdday + 1) % 7)


def whole_day(start, end):
    """Returns true if the given start, end range should be considered
    a whole-day range. This is so if the start time is 0:00:00 and the end
    time either 0:59:59 or 0:00:00 and if there is at least a diff
    erence of 23h 59m 59s / 86399 seconds between them.

    This is relevant for the calendar-display for now. This might very well be
    replaced again in the future when we introduce timezones.

    """

    assert start <= end, "The end needs to be equal or greater than the start"

    if isinstance(start, datetime_time):
        start = datetime(2000, 1, 1, start.hour, start.minute, start.second)

    if isinstance(end, datetime_time):
        end = datetime(2000, 1, 1, end.hour, end.minute, end.second)

    if (start.hour, start.minute, start.second) != (0, 0, 0):
        return False

    if (end.hour, end.minute, end.second) not in ((0, 0, 0), (23, 59, 59)):
        return False

    if (end - start).total_seconds() < 86399:
        return False

    return True


def context_path(context):
    """Returns the physical path for brains and objects alike."""

    if hasattr(context, 'getPath'):
        return context.getPath().split('/')  # same as getPhysicalPath
    else:
        return context.getPhysicalPath()


def portal_type_in_site(portal_type):
    catalog = getToolByName(getSite(), 'portal_catalog')
    results = catalog(portal_type=portal_type)

    return results


def portal_type_in_context(context, portal_type, depth=1):
    """Returns the given portal types _within_ the current context."""

    path = '/'.join(context_path(context))

    catalog = getToolByName(context, 'portal_catalog')
    results = catalog(
        portal_type=portal_type,
        path={'query': path, 'depth': depth}
    )
    return results


def portal_type_by_context(context, portal_type):
    """Returns the given portal_type _for_ the current context. Portal types
    for a context are required by traversing up the acquisition context.

    """

    # As this code is used for backend stuff that needs to be callable for
    # every user the idea is to do an unrestricted traverse.

    # Calling unrestrictedTraverse first seems to work with the supposedly
    # restricted catalog.

    path = context_path(context)
    context = context.aq_inner.unrestrictedTraverse(path)

    def traverse(context, portal_type):
        frames = portal_type_in_context(context, portal_type)
        if frames:
            return [f.getObject() for f in frames]
        else:
            if not hasattr(context, 'portal_type'):
                return []
            if context.portal_type == 'Plone Site':
                return []

            parent = context.aq_inner.aq_parent
            return traverse(parent, portal_type)

    return traverse(context, portal_type)


def align_date_to_day(date, direction):
    """ Aligns the given date to the beginning or end of the day, depending on
    the direction.

    E.g.
    2012-1-24 10:00 down -> 2012-1-24 00:00
    2012-1-24 10:00 up   -> 2012-1-24 23:59:59'999999

    """
    assert direction in ('up', 'down')

    aligned = (0, 0, 0, 0) if direction == 'down' else (23, 59, 59, 999999)

    if (date.hour, date.minute, date.second, date.microsecond) == aligned:
        return date

    if direction == 'down':
        return date.replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        date = date.replace(hour=0, minute=0, second=0, microsecond=0)
        return date + timedelta(days=1, microseconds=-1)


def align_range_to_day(start, end):
    assert start <= end, "{} - {} is an invalid range".format(start, end)

    return align_date_to_day(start, 'down'), align_date_to_day(end, 'up')


def display_date(start, end):
    """ Formates the date range given for display. """

    if end.microsecond != 999999:
        end -= timedelta(microseconds=1)

    if (start, end) == align_range_to_day(start, end):
        if start.date() == end.date():
            return api.portal.get_localized_time(start, long_format=False)
        else:
            return ' - '.join((
                api.portal.get_localized_time(start, long_format=False),
                api.portal.get_localized_time(end, long_format=False)
            ))

    end += timedelta(microseconds=1)

    if start.date() == end.date():
        return ' - '.join((
            api.portal.get_localized_time(start, long_format=True),
            api.portal.get_localized_time(end, time_only=True)
        ))
    else:
        return ' - '.join((
            api.portal.get_localized_time(start, long_format=True),
            api.portal.get_localized_time(end, long_format=True)
        ))


class United(object):
    """ Puts items added through 'append' into the same group as the last
    item which was appended, as long as the matchfn which is passed the last
    and the current item returns true.

    e.g.

    united = United(lambda last, current: last == current)

    united.append(1)
    -> united.groups => [[1]]

    united.append(1)
    -> united.groups => [[1, 1]]

    united.append(2)
    -> united.groups => [[1, 1], [2]]

    united.append(1)
    -> united.groups => [[1, 1], [2], [1]]

    """

    def __init__(self, matchfn):
        self.matchfn = matchfn
        self.groups = []
        self.new_group()

    def new_group(self):
        self.groups.append([])

    @property
    def current_group(self):
        return self.groups[-1]

    def append(self, item):
        if not self.current_group:
            self.current_group.append(item)
        else:
            if not self.matchfn(self.current_group[-1], item):
                self.new_group()

            self.current_group.append(item)


def unite(iterator, matchfn):
    """ Iterates through the given records and groups the records if two
    records passed to the match function result in True.

    e.g.

    unite([1, 1, 1, 2, 2], lambda last, current: last == current)

    -> [[1, 1, 1], [2, 2]]

    Note that for this to work the records need to be sorted, not unlike
    when using the collection's group function. See 'United' doc above to
    learn why.
    """

    united = United(matchfn)

    for item in iterator:
        united.append(item)

    return united.groups


def unite_dates(dates):
    """ Takes a list of start / end date tuples and compacts it by merging
    dates which can be considered an unbroken range.

    So given a list like this:

        [
            (01.01.2012 00:00, 02.01.2012 00:00),
            (02.01.2012 00:00, 03.01.2012 00:00),
            (05.01.2012 00:00, 06.01.2012 00:00)
        ]

    (Where the dates are actually datetime instances)

    We should get this result:

        [
            (01.01.2012 00:00, 03.01.2012 00:00),
            (05.01.2012 00:00, 06.01.2012 00:00)
        ]

    """
    sorted_dates = sorted(dates, key=lambda pair: pair[0])
    for group in unite(sorted_dates, lambda this, next: this[1] == next[0]):
        yield group[0][0], group[-1][1]


class cached_property(object):
    """A read-only @property that is only evaluated once. The value is cached
    on the object itself rather than the function or class; this should prevent
    memory leakage."""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result



import logging
import codecs
from os import path

logger = logging.getLogger('seantis.reservation')
folder = path.join(path.dirname(path.abspath(__file__)), 'emails')


def get_filename(key, language):
    filename = '%s.%s.txt' % (key, language)
    return path.join(folder, filename)


class MailTemplate(object):

    def __init__(self, template):
        self.templates = dict()
        self.key = template
        assert self.is_translated('en')

    def is_translated(self, language):
        return path.exists(get_filename(self.key, language))

    def load_language(self, language):
        with codecs.open(get_filename(self.key, language), "r", "utf-8") as f:
            self.templates[language] = self.parse_file(f)

        return self.templates[language]

    def get(self, language):

        if language in self.templates:
            return self.templates[language]

        if self.is_translated(language):
            return self.load_language(language)

        logger.warning(
            'Email template for language %s does not exist', language
        )

        return self.get('en')

    def get_subject(self, language):
        return self.get(language)[0]

    def get_body(self, language):
        return self.get(language)[1]

    def parse_file(self, template_file):

        subject = u''
        body = []

        for line in (l.strip('\n\t') for l in template_file):

            if line.startswith('#'):
                continue

            if not subject:
                subject = line
                continue

            if subject and not body and line.startswith('='):
                continue

            body.append(line)

        return unicode(subject), unicode('\n'.join(body))

keys = [
    'reservation_approved',
    'reservation_pending',
    'reservation_received',
    'reservation_denied',
    'reservation_revoked',
    'reservation_made',
    'reservation_time_changed'
]

templates = dict([(k, MailTemplate(k)) for k in keys])



from datetime import date, datetime, time
from five import grok
from plone.autoform.form import AutoExtensibleForm
from plone.directives import form
from plone.supermodel import model
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope import schema
from zope.security import checkPermission

from seantis.reservation import _
from seantis.reservation.utils import cached_property
from seantis.reservation.form import BaseForm
from seantis.reservation.resource import YourReservationsViewlet
from seantis.reservation.interfaces import IResourceBase, days as weekdays


class ISearchAndReserveForm(model.Schema):
    """ Search form for search & reserve view. """

    form.mode(timeframes='hidden')
    timeframes = schema.Text(
        title=_(u'Timeframes'),
        default=u'',
        required=False
    )

    recurrence_start = schema.Date(
        title=_(u"Start date"),
        required=True
    )

    recurrence_end = schema.Date(
        title=_(u"End date"),
        required=True
    )

    whole_day = schema.Bool(
        title=_(u"Whole day"),
        required=False,
        default=False
    )

    start_time = schema.Time(
        title=_(u"Start time"),
        required=False
    )

    end_time = schema.Time(
        title=_(u"End time"),
        required=False
    )

    form.widget(days=CheckBoxFieldWidget)
    days = schema.List(
        title=_(u"Days"),
        value_type=schema.Choice(vocabulary=weekdays),
        required=False
    )

    minspots = schema.Int(
        title=_(u"Spots"),
        required=False
    )

    available_only = schema.Bool(
        title=_(u"Available only"),
        required=False,
        default=True
    )


@form.default_value(field=ISearchAndReserveForm['recurrence_start'])
def start_default(data):
    return date.today()


@form.default_value(field=ISearchAndReserveForm['recurrence_end'])
def end_default(data):
    return date.today()


class SearchForm(BaseForm, AutoExtensibleForm, YourReservationsViewlet):
    permission = 'zope2.View'

    grok.context(IResourceBase)
    grok.require(permission)
    grok.name('search')

    ignoreContext = True

    template = grok.PageTemplateFile('templates/search.pt')
    schema = ISearchAndReserveForm

    enable_form_tabbing = False
    enable_unload_protection = False

    results = None
    searched = False

    start_time = None
    end_time = None

    # show the seantis.dir.facility viewlet if it's present
    show_facility_viewlet = True

    def update(self):
        super(SearchForm, self).update()
        self.widgets['timeframes'].value = self.context.json_timeframes()

    @property
    def available_actions(self):
        yield dict(name='search', title=_(u'Search'), css_class='context')

    @property
    def enable_removal(self):
        return checkPermission('cmf.ModifyPortalContent', self.context)

    @property
    def removal_url(self):
        return './remove-allocation?group={}'.format(
            '&group='.join(set(r.group.hex for r in self.results))
        )

    @cached_property
    def options(self):
        params = self.parameters

        if not params:
            return None

        options = {}

        options['days'] = tuple(d.weekday for d in params['days'])
        options['minspots'] = params['minspots'] or 0
        options['available_only'] = params['available_only']
        options['whole_day'] = params['whole_day'] and 'yes' or 'any'

        if options['whole_day'] == 'yes':
            start = datetime.combine(params['recurrence_start'], time(0, 0))
            end = datetime.combine(
                params['recurrence_end'], time(23, 59, 59, 999999)
            )
        else:
            self.start_time = params['start_time'] or time(0, 0)
            self.end_time = params['end_time'] or time(23, 59, 59, 999999)

            start = datetime.combine(
                params['recurrence_start'], self.start_time
            )
            end = datetime.combine(
                params['recurrence_end'], self.end_time
            )

        options['start'] = start
        options['end'] = end

        return options

    def handle_search(self):
        self.searched = True

        if not self.options:
            self.results = tuple()
        else:
            self.results = self.context.scheduler().search_allocations(
                **self.options
            )
